import random
import uuid
from typing import Tuple, List, Any

from omnisdk.omnitron.models import ProductStock, BatchRequest

from channel_app.core.commands import ChannelCommandInterface
from channel_app.core.data import BatchRequestResponseDto, ErrorReportDto
from channel_app.omnitron.constants import ResponseStatus


class SendUpdatedStocks(ChannelCommandInterface):
    """
    Daha once bildirilmiş stokların yeni güncellemelerini ilgili satış kanalına
    gönderebilmek için kullanılır.

    input olarak do_action'da
        objects -> List[ProductStock] tipinde kayıtları alır.
        batch_request -> BatchRequest tipinde kayıt alır. Ana işleme ait
            BatchRequest kaydı. Rapor üretmek için kullanılır.
        is_sync -> bool tipinde olan bu işlemin sonucunun hemen mi alınabileceği
            yoksa bu işlemin sonucunun asenkron mu öğrenilebileceğinin bilgisi

    List[BatchRequestResponseDto] -> Stok bildirimi yapildiginda sonucu hemen
        donuyorsa hazirlanir. Eğer bildirimin sonucu daha sonra kontrol edilecekse
        bu tip veri yerine None tipinde dönüş yapılır

    ErrorReportDto -> Rapor için üretilmiş  veri tipi. Hata olmasada uretilebilir.

    List[ProductStock] -> objects içerisinde geçilmiş veridir. Belki Monkey Patch
        yapılmış objeleri barındırıyor olabilir.

     :return: do_action çıktısı olarak
        (List[BatchRequestResponseDto], ErrorReportDto, List[ProductStock])
        listesi döner
    """
    param_sync = True

    def get_data(self) -> List[ProductStock]:
        product_stocks = self.objects
        return product_stocks

    def validated_data(self, data) -> object:
        return data

    def transform_data(self, data) -> object:
        return data

    def send_request(self, transformed_data) -> object:
        if not self.param_sync:
            response = self.__mocked_request(data=transformed_data)
        else:
            response = self.__mock_request_sync(data=transformed_data)

        return response

    def normalize_response(self, data, validated_data, transformed_data,
                           response) -> Tuple[List[BatchRequestResponseDto],
                                              ErrorReportDto, Any]:
        report = self.create_report(response)
        if not self.param_sync:
            remote_batch_id = response.get("remote_batch_request_id")
            self.batch_request.remote_batch_id = remote_batch_id
            return None, report, data
        else:
            response_data = []
            for row in response:
                response_data.append(BatchRequestResponseDto(
                    sku=row["sku"],
                    message=row["message"],
                    remote_id=row["remote_id"],
                    status=row["status"]
                ))

            response_data: List[BatchRequestResponseDto]
            return response_data, report, data

    def __mock_request_sync(self, data):
        result = []
        for row in data:
            obj = dict(
                sku=row["sku"],
                message=row["message"],
                remote_id=row["remote_id"],
                status=row["status"])
            result.append(obj)
        return result

    def __mocked_request(self, data):
        """
        Mock a request and response for the send operation to mimic actual channel data
        :param data:
        :return:
        """
        batch_id = str(uuid.uuid4())
        self.integration._sent_data[batch_id] = data
        return {"remote_batch_request_id": batch_id}


class SendInsertedStocks(SendUpdatedStocks):
    """
    Daha once bildirilmemiş stokları ilk defa bildirmesi için kullanılır.

    input olarak do_action'da
        objects -> List[ProductStock] tipinde kayıtları alır.
        batch_request -> BatchRequest tipinde kayıt alır. Ana işleme ait
            BatchRequest kaydı. Rapor üretmek için kullanılır.
        is_sync -> bool tipinde olan bu işlemin sonucunun hemen mi alınabileceği
            yoksa bu işlemin sonucunun asenkron mu öğrenilebileceğinin bilgisi

    List[BatchRequestResponseDto] -> Stok bildirimi yapildiginda sonucu hemen
        donuyorsa hazirlanir. Eğer bildirimin sonucu daha sonra kontrol edilecekse
        bu tip veri yerine None tipinde dönüş yapılır

    ErrorReportDto -> Rapor için üretilmiş  veri tipi. Hata olmasada uretilebilir.

    List[ProductStock] -> objects içerisinde geçilmiş veridir. Belki Monkey Patch
        yapılmış objeleri barındırıyor olabilir.

     :return: do_action çıktısı olarak
        (List[BatchRequestResponseDto], ErrorReportDto, List[ProductStock])
        listesi döner
    """
    pass


class CheckStocks(ChannelCommandInterface):
    """
    Bildirilmiş stokların durumlarını satış kanılana sorularak öğrenilmesi
    için kullanılır.

    input olarak do_action'da
        objects -> batch_request tipinde veri alır. Bu veri durumunu kontrol
            edilecek olan daha önce bildirim işleminde bulunmuş batch_request
            kaydıdır.
        batch_request -> BatchRequest tipinde kayıt alır. Ana işleme ait
            BatchRequest kaydı. Rapor üretmek için kullanılır.

    List[BatchRequestResponseDto] -> Stok bildirimi yapildiginda sonucu hemen
        donuyorsa hazirlanir. Eğer bildirimin sonucu daha sonra kontrol edilecekse
        bu tip veri yerine None tipinde dönüş yapılır

    ErrorReportDto -> Rapor için üretilmiş  veri tipi. Hata olmasada uretilebilir.

    BatchRequest -> objects içerisinde geçilmiş veridir. Belki Monkey Patch
        yapılmış objeleri barındırıyor olabilir.

     :return: do_action çıktısı olarak
        (List[BatchRequestResponseDto], ErrorReportDto, BatchRequest)
        listesi döner
    """

    def get_data(self) -> BatchRequest:
        batch_request = self.objects
        return batch_request

    def validated_data(self, data):
        return data

    def transform_data(self, data) -> object:
        return data

    def send_request(self, transformed_data) -> object:
        """
        Sends a post request to the channel client to check the stocks
        param validated_data:
        """

        response = self.__mocked_request(
            data=self.integration._sent_data[transformed_data.remote_batch_id],
            remote_batch_id=transformed_data.remote_batch_id)
        return response

    def normalize_response(self, data, validated_data, transformed_data,
                           response) -> Tuple[List[BatchRequestResponseDto],
                                              ErrorReportDto, Any]:
        response_data = []
        for row in response:
            obj = BatchRequestResponseDto(
                sku=row["sku"],
                message=row["message"],
                remote_id=row["remote_id"],
                status=row["status"])
            response_data.append(obj)
        report = self.create_report(response)
        return response_data, report, data

    def __mocked_request(self, data, remote_batch_id):
        """
        Mock a request and response for the send operation to mimic actual channel data
        :param data:
        :return:

        [{
            "status": "SUCCESS",
            "remote_id": "123a1",
            "sku": "1KBATC0197",
            "message": ""
         },]
        """

        response_data = []
        prefix = remote_batch_id[-8:]
        for index, item in enumerate(data):
            if random.random() < 0.8:
                response_item = {
                    'sku': item.sku,
                    'message': "",
                    'remote_id': "{}_{}".format(prefix, index),
                    'status': ResponseStatus.success
                }
            else:
                response_item = {
                    "status": ResponseStatus.fail,
                    "remote_id": None,
                    "sku": item.sku,
                    "message": "exception message"
                }
            response_data.append(response_item)
        return response_data



from dataclasses import asdict
from typing import List, Generator, Union

from requests import exceptions as requests_exceptions

from omnisdk.omnitron.models import (Customer, Address, CargoCompany, Order,
                                     BatchRequest, CancellationRequest)
from omnisdk.omnitron.endpoints import ChannelIntegrationActionEndpoint

from channel_app.core import settings
from channel_app.core.data import (BatchRequestResponseDto, 
                                   CancellationRequestDto, 
                                   ChannelCancellationRequestDto, 
                                   OmnitronCreateOrderDto, 
                                   OmnitronOrderDto,
                                   ChannelCreateOrderDto,
                                   ErrorReportDto,
                                   OrderBatchRequestResponseDto, 
                                   CancelOrderDto,
                                   ChannelUpdateOrderItemDto)
from channel_app.core.settings import OmnitronIntegration, ChannelIntegration
from channel_app.omnitron.batch_request import ClientBatchRequest
from channel_app.omnitron.constants import (BatchRequestStatus, ContentType, 
                                            FailedReasonType)
from channel_app.omnitron.exceptions import (CityException,
                                             TownshipException,
                                             DistrictException,
                                             CargoCompanyException,
                                             OrderException)


class OrderService(object):
    batch_service = ClientBatchRequest

    def fetch_and_create_order(self, is_success_log=True):
        with OmnitronIntegration(
                content_type=ContentType.order.value) as omnitron_integration:
            get_orders = ChannelIntegration().do_action(
                key='get_orders',
                batch_request=omnitron_integration.batch_request
            )

            get_orders: Generator
            order_batch_objects = []
            while True:
                try:
                    channel_create_order, report_list, _ = next(get_orders)
                except StopIteration:
                    break

                # tips
                channel_create_order: ChannelCreateOrderDto
                report_list: List[ErrorReportDto]
                for report in report_list:
                    if is_success_log or not report.is_ok:
                        report.error_code = \
                            f"{omnitron_integration.batch_request.local_batch_id}" \
                            f"-Channel-GetOrders_{channel_create_order.order.number}"
                        omnitron_integration.do_action(
                            key='create_error_report',
                            objects=report)

                order = self.create_order(omnitron_integration=omnitron_integration,
                                          channel_order=channel_create_order)
                if order and omnitron_integration.batch_request.objects:
                    order_batch_objects.extend(omnitron_integration.batch_request.objects)

            omnitron_integration.batch_request.objects = order_batch_objects
            try:
                self.batch_service(settings.OMNITRON_CHANNEL_ID).to_done(
                    batch_request=omnitron_integration.batch_request
                )
            except requests_exceptions.HTTPError as exc:
                if exc.response.status_code == 406 and "batch_request_status_100_1" in exc.response.text:
                    pass
                else:
                    raise exc

    def create_order(self, omnitron_integration: OmnitronIntegration,
                     channel_order: ChannelCreateOrderDto
                     ) -> Union[Order, None]:
        order = channel_order.order

        try:
            customer = omnitron_integration.do_action(
                key='get_or_create_customer',
                objects=order.customer)[0]
        except Exception:
            return

        customer: Customer
        try:
            shipping_address_data = {
                "customer": customer,
                "address": order.shipping_address
            }
            shipping_address = omnitron_integration.do_action(
                key='get_or_create_address',
                objects=shipping_address_data
            )[0]

            billing_address_data = {
                "customer": customer,
                "address": order.billing_address
            }
            billing_address = omnitron_integration.do_action(
                key='get_or_create_address',
                objects=billing_address_data
            )[0]
        except (CityException, TownshipException, DistrictException) as exc:
            omnitron_integration.do_action(
                key='create_address_error_report',
                object=exc)
            return

        except IndexError:
            return

        shipping_address: Address
        billing_address: Address
        try:
            cargo_company = omnitron_integration.do_action(
                key='get_cargo_company',
                objects=order.cargo_company
            )[0]
        except (CargoCompanyException, IndexError):
            return
        cargo_company: CargoCompany

        order_data = asdict(order)
        order_data["customer"] = customer.pk
        order_data["shipping_address"] = shipping_address.pk
        order_data["billing_address"] = billing_address.pk
        order_data["cargo_company"] = cargo_company.pk
        omnitron_order = OmnitronOrderDto(**order_data)
        create_order_dto = OmnitronCreateOrderDto(
            order=omnitron_order,
            order_item=channel_order.order_item)
        try:
            orders: List[Order] = omnitron_integration.do_action(
                key='create_order',
                objects=create_order_dto
            )
            order = orders[0]
        except (OrderException, IndexError):
            return

        return order

    def fetch_and_update_order_items(self, is_success_log=True):
        with OmnitronIntegration(
                content_type=ContentType.order.value) as omnitron_integration:
            get_updated_orders = ChannelIntegration().do_action(
                key='get_updated_order_items',
                batch_request=omnitron_integration.batch_request
            )
            get_updated_orders: Generator
            order_batch_objects = []
            while True:
                try:
                    channel_update_order, report_list, _ = next(get_updated_orders)
                except StopIteration:
                    break

                # tips
                channel_update_order: ChannelUpdateOrderItemDto
                report_list: List[ErrorReportDto]
                for report in report_list:
                    if report and (is_success_log or not report.is_ok):
                        report.error_code = \
                            f"{omnitron_integration.batch_request.local_batch_id}" \
                            f"_GetUpdatedOrders_{channel_update_order.remote_id}"
                        omnitron_integration.do_action(
                            key='create_error_report',
                            objects=report)

                omnitron_integration.do_action(
                    key='update_order_items', objects=channel_update_order)

            omnitron_integration.batch_request.objects = order_batch_objects

            self.batch_service(settings.OMNITRON_CHANNEL_ID).to_done(
                batch_request=omnitron_integration.batch_request
            )


    def update_orders(self, is_sync=True, is_success_log=True,
                      add_order_items=False):
        with OmnitronIntegration(
                content_type=ContentType.order.value) as omnitron_integration:
            orders = omnitron_integration.do_action(key='get_orders')
            orders: List[Order]
            first_order_count = len(orders)
            if add_order_items:
                orders = orders and omnitron_integration.do_action(
                    key='get_order_items_with_order', objects=orders)

            if not orders:
                if first_order_count:
                    omnitron_integration.batch_request.objects = None
                    self.batch_service(omnitron_integration.channel_id).to_fail(
                        omnitron_integration.batch_request
                    )
                return

            response_data, reports, data = ChannelIntegration().do_action(
                key='send_updated_orders',
                objects=orders,
                batch_request=omnitron_integration.batch_request,
                is_sync=True)

            # tips
            response_data: List[OrderBatchRequestResponseDto]
            reports: List[ErrorReportDto]
            data: List[Order]

            if not is_sync:
                if reports[0].is_ok:
                    self.batch_service(
                        settings.OMNITRON_CHANNEL_ID).to_sent_to_remote(
                        batch_request=omnitron_integration.batch_request)
                else:
                    is_sync = True

            if reports and (is_success_log or not reports[0].is_ok):
                for report in reports:
                    omnitron_integration.do_action(
                        key='create_error_report',
                        objects=report)

            if is_sync:
                omnitron_integration.do_action(
                    key='process_order_batch_requests',
                    objects=response_data)

    def get_order_batch_requests(self, is_success_log=False):
        with OmnitronIntegration(create_batch=False) as omnitron_integration:
            batch_request_data = omnitron_integration.do_action(
                'get_batch_requests',
                params={
                    "status": ["sent_to_remote", "ongoing"],
                    "content_type": ContentType.order.value})

            # tips
            batch_request_data: List[BatchRequest]

            for batch_request in batch_request_data:
                response_data, report, data = ChannelIntegration().do_action(
                    key='check_orders',
                    objects=batch_request)

                # tips
                response_data: List[OrderBatchRequestResponseDto]
                report: ErrorReportDto
                data: BatchRequest

                if report and (is_success_log or not report.is_ok):
                    omnitron_integration.do_action(
                        key='create_error_report',
                        objects=report)

                if response_data:
                    omnitron_integration.batch_request = batch_request_data
                    omnitron_integration.do_action(
                        key='process_order_batch_requests',
                        objects=response_data)

    def fetch_and_create_cancel(self, is_success_log=True):
        with OmnitronIntegration(
                content_type=ContentType.order.value) as omnitron_integration:
            get_cancelled_order = ChannelIntegration().do_action(
                key='get_cancelled_orders',
                batch_request=omnitron_integration.batch_request
            )
            get_cancelled_order: Generator

            while True:
                try:
                    cancel_order_dto, report, _ = next(get_cancelled_order)
                except StopIteration:
                    break

                # tips
                cancel_order_dto: CancelOrderDto
                report: ErrorReportDto

                if report and (is_success_log or not report.is_ok):
                    omnitron_integration.do_action(
                        key='create_error_report',
                        objects=report)

                self.create_cancel(omnitron_integration=omnitron_integration,
                                   cancel_order_dto=cancel_order_dto)

    def create_cancel(self, omnitron_integration: OmnitronIntegration,
                      cancel_order_dto: CancelOrderDto):
        success_datas: List[Order] = omnitron_integration.do_action(
            key="create_order_cancel",
            objects=cancel_order_dto)
        try:
            success_data = success_datas[0]
            return success_data
        except IndexError:
            return

    def fetch_and_create_cancellation_requests(self, is_success_log=True):
        with OmnitronIntegration(
                content_type=ContentType.cancellation_request.value) as omnitron_integration:
            get_cancellation_requests = ChannelIntegration().do_action(
                key='get_cancellation_requests',
                batch_request=omnitron_integration.batch_request)
            get_cancellation_requests: Generator
            while True:
                try:
                    cancellation_request, report_list, _ = next(get_cancellation_requests)
                except StopIteration:
                    break

                # tips
                cancellation_request: CancellationRequestDto
                report_list: List[ErrorReportDto]
                for report in report_list:
                    if report and (is_success_log or not report.is_ok):
                        report.error_code = \
                            f"{omnitron_integration.batch_request.local_batch_id}" \
                            f"-Channel-GetCancellationRequests_{cancellation_request.order_item}"                        
                        omnitron_integration.do_action(
                            key='create_error_report',
                            objects=report)
                
                # omnitron integration do action create_cancellation_request
                cancellation_request_response = omnitron_integration.do_action(
                    key='create_cancellation_requests',
                    objects=cancellation_request)
            try:
                self.batch_service(settings.OMNITRON_CHANNEL_ID).to_done(
                    batch_request=omnitron_integration.batch_request
                )
            except requests_exceptions.HTTPError as exc:
                if exc.response.status_code == 406 and "batch_request_status_100_1" in exc.response.text:
                    pass
                else:
                    raise exc                
                
                            
    

    def update_cancellation_requests(self, is_success_log=True):
        with OmnitronIntegration(
                content_type=ContentType.cancellation_request.value) as omnitron_integration:
            cancellation_requests = omnitron_integration.do_action(
                key='get_cancellation_requests_update', objects={})
            cancellation_requests: List[CancellationRequest]

            if not cancellation_requests:
                omnitron_integration.batch_request.objects = None
                self.batch_service(omnitron_integration.channel_id).to_fail(
                    omnitron_integration.batch_request)
                return
            
            batch_request_object_list = []
            
            for cancellation_request in cancellation_requests:
                remote_order_item = self.get_channel_order_item(
                    omnitron_integration, cancellation_request.order_item)
                remote_reason = self.get_channel_reason(
                    omnitron_integration, cancellation_request.reason)
                remote_cancellation_request = self.get_channel_cancellation_request(
                    omnitron_integration, cancellation_request.id)
                
                channel_cancellation_request = ChannelCancellationRequestDto(
                    cancellation_type=cancellation_request.cancellation_type,
                    status=cancellation_request.status,
                    order_item=remote_order_item,
                    reason=remote_reason,
                    description=cancellation_request.description,
                    remote_id=remote_cancellation_request,
                )
                response_data, reports, data = ChannelIntegration().do_action(
                    key='update_cancellation_request',
                    objects=channel_cancellation_request,
                    batch_request=omnitron_integration.batch_request,
                    is_sync=True)

                # tips
                response_data: List[BatchRequestResponseDto]
                reports: List[ErrorReportDto]
                data: List[ChannelCancellationRequestDto]

                if reports and (is_success_log or not reports[0].is_ok):
                    for report in reports:
                        omnitron_integration.do_action(
                            key='create_error_report',
                            objects=report)

                if response_data:
                    failed_reason_type = None
                else:
                    failed_reason_type = FailedReasonType.remote.value
                
                batch_request_object_dto = dict(
                    pk=cancellation_request.id,
                    version_date=cancellation_request.modified_date,
                    content_type=ContentType.cancellation_request.value,
                    failed_reason_type=failed_reason_type,
                    remote_id=remote_cancellation_request)
                
                batch_request_object_list.append(batch_request_object_dto)
                   
            status = BatchRequestStatus.fail.value
            if any([br["failed_reason_type"] is None for br in batch_request_object_list]):
                status = BatchRequestStatus.done.value

            omnitron_integration.batch_request.objects = batch_request_object_list
            service = self.batch_service(omnitron_integration.channel_id)
            
            if status == BatchRequestStatus.done.value:
                service.to_done(batch_request=omnitron_integration.batch_request)
            else:
                service.to_fail(batch_request=omnitron_integration.batch_request)

    def get_channel_order_item(self, omnitron_integration, order_item):
        channel_id = omnitron_integration.channel_id
        end_point = ChannelIntegrationActionEndpoint(
            channel_id=channel_id)
        params = {
            "channel": channel_id,
            "content_type_name": ContentType.order_item.value,
            "object_id": order_item
        }
        integration_action = end_point.list(params=params)
        if not integration_action:
            return Exception("Order item remote id not found: {}".format(order_item))
        if len(integration_action) > 1:
            return Exception("Multiple order item remote id found: {}".format(order_item))
        return integration_action[0].remote_id
    
    def get_channel_reason(self, omnitron_integration, omnitron_reason):
        configuration = omnitron_integration.channel.conf
        # configuration icinde yer alan reason_mapping anahtari bir json objesi,
        # key channel_reason value omnitron_reason olan bir dict objesinden
        # channel_reason'u döndürür
        for channel_reason, omni_reason in configuration.get("reason_mapping", {}):
            if omnitron_reason == omni_reason:
                return channel_reason
        
        return "10"
    
    def get_channel_cancellation_request(self, omnitron_integration, 
                                   omnitron_cancel_request_pk):
        channel_id = omnitron_integration.channel_id
        end_point = ChannelIntegrationActionEndpoint(
            channel_id=channel_id)
        params = {
            "channel": channel_id,
            "content_type_name": ContentType.cancellation_request.value,
            "object_id": omnitron_cancel_request_pk
        }
        integration_action = end_point.list(params=params)
        if not integration_action:
            return Exception("Cancellation request remote id not found: {}".format(
                omnitron_cancel_request_pk))
        if len(integration_action) > 1:
            return Exception("Multiple cancellation request remote id found: {}".format(
                omnitron_cancel_request_pk))
        return integration_action[0].remote_id




from typing import List

from omnisdk.omnitron.models import BatchRequest, ProductImage

from channel_app.core import settings
from channel_app.core.data import BatchRequestResponseDto, ErrorReportDto
from channel_app.core.settings import OmnitronIntegration, ChannelIntegration
from channel_app.omnitron.batch_request import ClientBatchRequest
from channel_app.omnitron.constants import ContentType


class ImageService(object):
    batch_service = ClientBatchRequest

    def update_product_images(self, is_sync=True, is_success_log=True,
                              add_product_objects=False):
        with OmnitronIntegration(
                content_type=ContentType.product_image.value) as omnitron_integration:
            product_images = omnitron_integration.do_action(
                key='get_updated_images')
            first_product_image_count = len(product_images)
            if add_product_objects:
                product_images = product_images and omnitron_integration.do_action(
                    key='get_product_objects', objects=product_images)

            if not product_images:
                if first_product_image_count:
                    omnitron_integration.batch_request.objects = None
                    self.batch_service(omnitron_integration.channel_id).to_fail(
                        omnitron_integration.batch_request
                    )
                return

            product_images: List[ProductImage]

            response_data, reports, data = ChannelIntegration().do_action(
                key='send_updated_images',
                objects=product_images,
                batch_request=omnitron_integration.batch_request,
                is_sync=is_sync)

            # tips
            response_data: List[BatchRequestResponseDto]
            reports: List[ErrorReportDto]
            data: List[ProductImage]

            if not is_sync:
                if reports[0].is_ok:
                    self.batch_service(
                        settings.OMNITRON_CHANNEL_ID).to_sent_to_remote(
                        batch_request=omnitron_integration.batch_request)
                else:
                    is_sync = True

            if reports and (is_success_log or not reports[0].is_ok):
                for report in reports:
                    omnitron_integration.do_action(
                        key='create_error_report',
                        objects=report)

            if is_sync:
                omnitron_integration.do_action(
                    key='process_image_batch_requests',
                    objects=response_data)

    def insert_product_images(self, is_sync=True, is_success_log=True,
                              add_product_objects=False):
        with OmnitronIntegration(
                content_type=ContentType.product_image.value) as omnitron_integration:
            product_images = omnitron_integration.do_action(
                key='get_inserted_images')
            first_product_image_count = len(product_images)

            if add_product_objects:
                product_images = product_images and omnitron_integration.do_action(
                    key='get_product_objects', objects=product_images)

            if not product_images:
                if first_product_image_count:
                    omnitron_integration.batch_request.objects = None
                    self.batch_service(omnitron_integration.channel_id).to_fail(
                        omnitron_integration.batch_request
                    )
                return

            product_images: List[ProductImage]

            response_data, reports, data = ChannelIntegration().do_action(
                key='send_inserted_images',
                objects=product_images,
                batch_request=omnitron_integration.batch_request,
                is_sync=is_sync)

            # tips
            response_data: List[BatchRequestResponseDto]
            reports: List[ErrorReportDto]
            data: List[ProductImage]

            if not is_sync:
                if reports[0].is_ok:
                    self.batch_service(
                        settings.OMNITRON_CHANNEL_ID).to_sent_to_remote(
                        batch_request=omnitron_integration.batch_request)
                else:
                    is_sync = True

            if reports and (is_success_log or not reports[0].is_ok):
                for report in reports:
                    omnitron_integration.do_action(
                        key='create_error_report',
                        objects=report)

            if is_sync:
                omnitron_integration.do_action(
                    key='process_image_batch_requests',
                    objects=response_data)

    def get_image_batch_requests(self, is_success_log=True):
        with OmnitronIntegration(create_batch=False) as omnitron_integration:
            batch_request_data = omnitron_integration.do_action(
                'get_batch_requests',
                params={
                    "status": ["sent_to_remote", "ongoing"],
                    "content_type": ContentType.product_image.value})
            # tips
            batch_request_data: List[BatchRequest]

            for batch_request in batch_request_data:
                response_data, reports, data = ChannelIntegration().do_action(
                    key='check_images',
                    objects=batch_request,
                    batch_request=batch_request
                )

                # tips
                response_data: List[BatchRequestResponseDto]
                reports: List[ErrorReportDto]
                data: BatchRequest

                if reports and (is_success_log or not reports[0].is_ok):
                    for report in reports:
                        omnitron_integration.do_action(
                            key='create_error_report',
                            objects=report)

                if response_data:
                    omnitron_integration.batch_request = batch_request
                    omnitron_integration.do_action(
                        key='process_image_batch_requests',
                        objects=response_data)



from typing import Tuple, Any, List

from omnisdk.omnitron.models import IntegrationAction

from channel_app.core.commands import ChannelCommandInterface
from channel_app.core.data import (CategoryTreeDto, CategoryNodeDto,
                                   CategoryDto,
                                   CategoryAttributeDto,
                                   CategoryAttributeValueDto, ErrorReportDto,
                                   ChannelConfSchemaField, AttributeDto,
                                   AttributeValueDto)
from channel_app.core.utilities import mock_response_decorator
from channel_app.omnitron.constants import ChannelConfSchemaDataTypes


class GetCategoryTreeAndNodes(ChannelCommandInterface):
    """
    Category trees and categories represent an easy way to find, list, filter
    products. Tree representation of the sales channel must be fetched and
    all products we want to sell must be assigned to a category.

    This command fetches all categories belonging to a sales channels and their
    tree structure.

    input olarak do_action'da
        objects -> None tipinde kayıt alır. Yani veri almaz
        batch_request -> BatchRequest tipinde kayıt alır. Ana işleme ait
            BatchRequest kaydı. Rapor üretmek için kullanılır.

    CategoryTreeDto -> Kategori ağacı ve buna bağlı dalları barındıran veri tipi.

    ErrorReportDto -> Rapor için üretilmiş  veri tipi. Hata olmasada uretilebilir.

    """

    def get_data(self):

        data = self.objects
        return data

    def validated_data(self, data) -> object:
        return data

    def transform_data(self, data) -> object:
        return data

    def send_request(self, transformed_data) -> object:
        response = self.__mocked_request(data=transformed_data)
        return response

    def normalize_response(self, data, validated_data, transformed_data,
                           response) -> Tuple[
        CategoryTreeDto, List[ErrorReportDto], Any]:
        """
        Convert channel category tree to the format OmnitronIntegration
        requires
        """
        report = self.create_report(response)
        node_dict = {}

        stack = []
        current = {
            "id": None,
            "name": "root",
            "parentId": None,
            "subCategories": response.json()["categories"]
        }
        # TODO document parent of top level nodes should be None
        # TODO document id and parent_id of root node must be None
        while True:
            if current:
                category_node = CategoryNodeDto(
                    parent=node_dict[current["parentId"]] if current[
                        "parentId"] else None,
                    remote_id=current["id"],
                    children=[],
                    name=current["name"]
                )
                node_dict[current["id"]] = category_node
                for child in current["subCategories"]:
                    stack.append(child)
                current = None
            elif stack:
                current = stack.pop()
            else:
                break
        category_tree = CategoryTreeDto(root=node_dict[None])
        node_dict.pop(None)
        for node_remote_id, node in node_dict.items():
            if node.parent is not None:
                node_dict[node.parent.remote_id].children.append(node)
            else:
                category_tree.root.children.append(node)

        return category_tree, report, data

    @mock_response_decorator
    def __mocked_request(self, data):
        """
        Mock a request and response for the send operation to mimic actual channel data

        """
        return {
            "categories": [
                {
                    "id": 403,
                    "name": "Ayakkabı",
                    "parentId": None,
                    "subCategories": [
                        {
                            "id": 420,
                            "name": "Spor Ayakkabı",
                            "parentId": 403,
                            "subCategories": [
                                {
                                    "id": 421,
                                    "name": "Basketbol Ayakkabısı",
                                    "parentId": 420,
                                    "subCategories": []
                                },
                                {
                                    "id": 422,
                                    "name": "Fitness Ayakkabısı",
                                    "parentId": 420,
                                    "subCategories": []
                                },
                                {
                                    "id": 426,
                                    "name": "Halı Saha Ayakkabısı&Krampon",
                                    "parentId": 420,
                                    "subCategories": []
                                },
                                {
                                    "id": 425,
                                    "name": "Koşu & Antrenman Ayakkabısı",
                                    "parentId": 420,
                                    "subCategories": [
                                        {
                                            "id": 999,
                                            "name": "Test Sub Category",
                                            "parentId": 425,
                                            "subCategories": []
                                        }
                                    ]
                                }
                            ]
                        }
                    ]
                }
            ]
        }


class GetCategoryAttributes(ChannelCommandInterface):
    """
    Along with classification, categories define a set of common attributes which makes
    sense only for that category and not others. To manage this process the attributes
    are fetched from the channel and certain validations are done to ensure product is
    ready to be sold before sending to the channel.

    This command fetches category attributes of a given category id.
    """

    def get_data(self):
        if self.objects:
            assert isinstance(self.objects, IntegrationAction)
        data = self.objects
        return data

    def validated_data(self, data) -> object:
        return data

    def transform_data(self, data) -> object:
        return data

    def send_request(self, transformed_data) -> object:
        response = self.__mocked_request(data=transformed_data)
        return response

    def normalize_response(self, data, validated_data, transformed_data,
                           response) -> Tuple[CategoryDto, ErrorReportDto, Any]:
        """
        Convert channel category tree to the format OmnitronIntegration
        requires
        """
        report = self.create_report(response)
        response = response.json()
        category = CategoryDto(remote_id=response["id"],
                               name=response["name"],
                               attributes=[])

        for channel_attribute in response["categoryAttributes"]:
            attribute = CategoryAttributeDto(
                remote_id=channel_attribute["attribute"]["id"],
                name=channel_attribute["attribute"]["name"],
                allow_custom_value=channel_attribute["allowCustom"],
                required=channel_attribute["required"],
                variant=channel_attribute["varianter"],
                values=[]
            )
            for channel_attribute_value in channel_attribute["attributeValues"]:
                attribute_value = CategoryAttributeValueDto(
                    remote_id=channel_attribute_value["id"],
                    name=channel_attribute_value["name"])
                attribute.values.append(attribute_value)
            category.attributes.append(attribute)
        return category, report, data

    @mock_response_decorator
    def __mocked_request(self, data):
        """
        Mock a request and response for the send operation to mimic actual channel data
        """
        return {
            "id": 123123124214,
            "name": "Sweatshirt",
            "displayName": "Sweatshirt",
            "categoryAttributes": [
                {
                    "allowCustom": False,
                    "attribute": {
                        "id": 2,
                        "name": "Boy / Ölçü"
                    },
                    "attributeValues": [
                        {
                            "id": 10,
                            "name": "Kısa"
                        },
                        {
                            "id": 16,
                            "name": "Uzun"
                        },
                        {
                            "id": 572,
                            "name": "Orta"
                        },
                        {
                            "id": 575,
                            "name": "Battal Boy"
                        }
                    ],
                    "categoryId": 601,
                    "required": False,
                    "varianter": False,
                    "slicer": False
                },
                {
                    "allowCustom": False,
                    "attribute": {
                        "id": 12,
                        "name": "Kol Tipi"
                    },
                    "attributeValues": [
                        {
                            "id": 62,
                            "name": "Askılı"
                        },
                        {
                            "id": 64,
                            "name": "Kısa Kol"
                        },
                        {
                            "id": 65,
                            "name": "Kolsuz"
                        },
                        {
                            "id": 67,
                            "name": "Tek Kol"
                        },
                        {
                            "id": 68,
                            "name": "Uzun"
                        },
                        {
                            "id": 69,
                            "name": "Uzun Kol"
                        },
                        {
                            "id": 622,
                            "name": "Geniş / Yarasa"
                        },
                        {
                            "id": 2841,
                            "name": "Volanlı"
                        },
                        {
                            "id": 2842,
                            "name": "Omzu açık"
                        },
                        {
                            "id": 2857,
                            "name": "Balon"
                        },
                        {
                            "id": 4020,
                            "name": "Ay"
                        },
                        {
                            "id": 7118,
                            "name": "Karpuz Kol"
                        },
                        {
                            "id": 23532,
                            "name": "Reglan Kol"
                        },
                        {
                            "id": 193959,
                            "name": "Mendil"
                        },
                        {
                            "id": 256024,
                            "name": "Yarasa Kol"
                        },
                        {
                            "id": 666277,
                            "name": "Takma Kol"
                        },
                        {
                            "id": 902440,
                            "name": "Ay Kol"
                        },
                        {
                            "id": 902442,
                            "name": "Standart Kol"
                        },
                        {
                            "id": 902443,
                            "name": "Düşük Kol"
                        },
                        {
                            "id": 902444,
                            "name": "Mendil Kol"
                        },
                        {
                            "id": 902445,
                            "name": "Balon Kol"
                        },
                        {
                            "id": 997508,
                            "name": "Büzgülü/Fırfırlı Kol"
                        },
                        {
                            "id": 997516,
                            "name": "İspanyol Kol"
                        }
                    ],
                    "categoryId": 601,
                    "required": False,
                    "varianter": False,
                    "slicer": False
                },
                {
                    "allowCustom": False,
                    "attribute": {
                        "id": 14,
                        "name": "Materyal"
                    },
                    "attributeValues": [
                        {
                            "id": 77,
                            "name": "Akrilik"
                        },
                        {
                            "id": 79,
                            "name": "Bambu"
                        },
                        {
                            "id": 91,
                            "name": "Kumaş"
                        },
                        {
                            "id": 99,
                            "name": "Polyester"
                        },
                        {
                            "id": 105,
                            "name": "Suni Deri"
                        },
                        {
                            "id": 195,
                            "name": "Kadife"
                        },
                        {
                            "id": 627,
                            "name": "Polar"
                        },
                        {
                            "id": 645,
                            "name": "Dalgıç Kumaş"
                        },
                        {
                            "id": 646,
                            "name": "Şerpa"
                        },
                        {
                            "id": 654,
                            "name": "Poliamid"
                        },
                        {
                            "id": 661,
                            "name": "PVC"
                        },
                        {
                            "id": 674,
                            "name": "Modal"
                        },
                        {
                            "id": 675,
                            "name": "Saten"
                        },
                        {
                            "id": 678,
                            "name": "Likra"
                        },
                        {
                            "id": 679,
                            "name": "Zincir"
                        },
                        {
                            "id": 681,
                            "name": "Pamuklu"
                        },
                        {
                            "id": 690,
                            "name": "Viskon"
                        },
                        {
                            "id": 710,
                            "name": "Keten"
                        },
                        {
                            "id": 713,
                            "name": "Elyaf"
                        },
                        {
                            "id": 727,
                            "name": "Pamuk-Polyester"
                        },
                        {
                            "id": 7119,
                            "name": "Lyocell"
                        },
                        {
                            "id": 9328,
                            "name": "Naylon"
                        },
                        {
                            "id": 9329,
                            "name": "Liyosel"
                        },
                        {
                            "id": 9330,
                            "name": "Poliüretan"
                        },
                        {
                            "id": 9331,
                            "name": "Asetat"
                        },
                        {
                            "id": 12277,
                            "name": "Elastan"
                        },
                        {
                            "id": 12278,
                            "name": "Fibre"
                        },
                        {
                            "id": 12279,
                            "name": "Linen"
                        },
                        {
                            "id": 12280,
                            "name": "Lureks"
                        },
                        {
                            "id": 12281,
                            "name": "Metalik İplik"
                        },
                        {
                            "id": 12282,
                            "name": "Mikro Polyester"
                        },
                        {
                            "id": 12283,
                            "name": "Poliakrilik"
                        },
                        {
                            "id": 12284,
                            "name": "Spandeks"
                        },
                        {
                            "id": 12285,
                            "name": "Viskoz"
                        },
                        {
                            "id": 12286,
                            "name": "Yün"
                        },
                        {
                            "id": 16668,
                            "name": "%100 Viskon"
                        },
                        {
                            "id": 16826,
                            "name": "%100 Polyester"
                        },
                        {
                            "id": 19513,
                            "name": "Pamuk-Elastan"
                        },
                        {
                            "id": 22135,
                            "name": "Pamuk-Polyester-Elastan"
                        },
                        {
                            "id": 23674,
                            "name": "Elastan-Viskon-Polyester"
                        },
                        {
                            "id": 137105,
                            "name": "Yün - Akrilik - Polyester"
                        },
                        {
                            "id": 144229,
                            "name": "%100 Organik Pamuk"
                        },
                        {
                            "id": 144666,
                            "name": "Pamuk-Viskon-Polyester"
                        },
                        {
                            "id": 165153,
                            "name": "Polietilen"
                        },
                        {
                            "id": 176093,
                            "name": "Pamuk - Polyamid"
                        },
                        {
                            "id": 194604,
                            "name": "Pamuk - Polietilen"
                        },
                        {
                            "id": 212262,
                            "name": "%100 Pamuk"
                        },
                        {
                            "id": 997542,
                            "name": "Polyamid"
                        },
                        {
                            "id": 997984,
                            "name": "Viskon Karışımlı"
                        },
                        {
                            "id": 997987,
                            "name": "Polyester Karışımlı"
                        },
                        {
                            "id": 997994,
                            "name": "Polyamid Karışımlı"
                        },
                        {
                            "id": 998003,
                            "name": "Yün Karışımlı"
                        },
                        {
                            "id": 998008,
                            "name": "Pamuk Karışımlı"
                        },
                        {
                            "id": 998013,
                            "name": "Pamuk Polyester"
                        },
                        {
                            "id": 998020,
                            "name": "Keten Karışımlı"
                        },
                        {
                            "id": 1005632,
                            "name": "Lyocell Karışımlı"
                        },
                        {
                            "id": 1005633,
                            "name": "Keten Görünümlü"
                        },
                        {
                            "id": 1178581,
                            "name": "Pamuk Elastan"
                        },
                        {
                            "id": 1178593,
                            "name": "Polyester Elastan"
                        }
                    ],
                    "categoryId": 601,
                    "required": False,
                    "varianter": False,
                    "slicer": False
                },
                {
                    "allowCustom": False,
                    "attribute": {
                        "id": 18,
                        "name": "Parça Sayısı"
                    },
                    "attributeValues": [
                        {
                            "id": 314394,
                            "name": "3"
                        },
                        {
                            "id": 314396,
                            "name": "1"
                        },
                        {
                            "id": 314398,
                            "name": "2"
                        },
                        {
                            "id": 314400,
                            "name": "5"
                        },
                        {
                            "id": 314401,
                            "name": "4"
                        },
                        {
                            "id": 417420,
                            "name": "6"
                        }
                    ],
                    "categoryId": 601,
                    "required": False,
                    "varianter": False,
                    "slicer": False
                },
                {
                    "allowCustom": False,
                    "attribute": {
                        "id": 22,
                        "name": "Yaka Tipi"
                    },
                    "attributeValues": [
                        {
                            "id": 175,
                            "name": "Balıkçı Yaka"
                        },
                        {
                            "id": 176,
                            "name": "Bebe Yaka"
                        },
                        {
                            "id": 177,
                            "name": "Bisiklet Yaka"
                        },
                        {
                            "id": 178,
                            "name": "Degaje Yaka"
                        },
                        {
                            "id": 179,
                            "name": "Dik Yaka"
                        },
                        {
                            "id": 180,
                            "name": "Hakim Yaka"
                        },
                        {
                            "id": 181,
                            "name": "Kare Yaka"
                        },
                        {
                            "id": 182,
                            "name": "Kayık Yaka"
                        },
                        {
                            "id": 184,
                            "name": "Kruvaze Yaka"
                        },
                        {
                            "id": 186,
                            "name": "Kalp Yaka"
                        },
                        {
                            "id": 188,
                            "name": "Straplez"
                        },
                        {
                            "id": 189,
                            "name": "Şal Yaka"
                        },
                        {
                            "id": 190,
                            "name": "U Yaka"
                        },
                        {
                            "id": 191,
                            "name": "V-Yaka"
                        },
                        {
                            "id": 192,
                            "name": "Yuvarlak Yaka"
                        },
                        {
                            "id": 779,
                            "name": "Marine"
                        },
                        {
                            "id": 780,
                            "name": "Asimetrik Yaka"
                        },
                        {
                            "id": 781,
                            "name": "Boyundan Bağlamalı"
                        },
                        {
                            "id": 782,
                            "name": "Carmen Yaka"
                        },
                        {
                            "id": 783,
                            "name": "Halter Yaka"
                        },
                        {
                            "id": 784,
                            "name": "V Yaka"
                        },
                        {
                            "id": 785,
                            "name": "Choker Yaka"
                        },
                        {
                            "id": 786,
                            "name": "Boğazlı"
                        },
                        {
                            "id": 788,
                            "name": "Klasik"
                        },
                        {
                            "id": 789,
                            "name": "Kolej Yaka"
                        },
                        {
                            "id": 792,
                            "name": "Yarım Balıkçı Yaka"
                        },
                        {
                            "id": 793,
                            "name": "Ata Yaka"
                        },
                        {
                            "id": 795,
                            "name": "Gizli Düğmeli Yaka"
                        },
                        {
                            "id": 796,
                            "name": "İtalyan Yaka"
                        },
                        {
                            "id": 797,
                            "name": "Yarım İtalyan Yaka"
                        },
                        {
                            "id": 798,
                            "name": "Smokin"
                        },
                        {
                            "id": 799,
                            "name": "Beyzbol Yaka"
                        },
                        {
                            "id": 800,
                            "name": "Resort Yaka"
                        },
                        {
                            "id": 801,
                            "name": "Çentik Yaka"
                        },
                        {
                            "id": 802,
                            "name": "Sivri Yaka"
                        },
                        {
                            "id": 803,
                            "name": "Oyuk Yaka"
                        },
                        {
                            "id": 804,
                            "name": "Düğmeli"
                        },
                        {
                            "id": 805,
                            "name": "Fermuarlı"
                        },
                        {
                            "id": 806,
                            "name": "Kapamasız"
                        },
                        {
                            "id": 807,
                            "name": "Gömlek Yaka"
                        },
                        {
                            "id": 808,
                            "name": "Kapüşonlu"
                        },
                        {
                            "id": 2838,
                            "name": "Devrik Yaka"
                        },
                        {
                            "id": 2839,
                            "name": "İspanyol Yaka"
                        },
                        {
                            "id": 2840,
                            "name": "Dantel Yaka"
                        },
                        {
                            "id": 2845,
                            "name": "Madonna Yaka"
                        },
                        {
                            "id": 2846,
                            "name": "Havuz Yaka"
                        },
                        {
                            "id": 2847,
                            "name": "İspanyol"
                        },
                        {
                            "id": 2852,
                            "name": "Polo Yaka"
                        },
                        {
                            "id": 2854,
                            "name": "O Yaka"
                        },
                        {
                            "id": 2864,
                            "name": "Sıfır Yaka"
                        },
                        {
                            "id": 2870,
                            "name": "Geniş Yaka"
                        },
                        {
                            "id": 22140,
                            "name": "Balerin Yaka"
                        },
                        {
                            "id": 22141,
                            "name": "Alttan Biritli Yaka"
                        },
                        {
                            "id": 22142,
                            "name": "Mono Yaka"
                        },
                        {
                            "id": 22143,
                            "name": "Kırlangıç Yaka"
                        },
                        {
                            "id": 22144,
                            "name": "Tulum Yaka"
                        },
                        {
                            "id": 23359,
                            "name": "Pis Yaka"
                        },
                        {
                            "id": 175940,
                            "name": "Ceket Yaka"
                        },
                        {
                            "id": 175941,
                            "name": "Erkek Yaka"
                        },
                        {
                            "id": 175942,
                            "name": "Fular Yaka"
                        },
                        {
                            "id": 194557,
                            "name": "Bomber Yaka"
                        },
                        {
                            "id": 256019,
                            "name": "Kruvaze"
                        },
                        {
                            "id": 260637,
                            "name": "Apaş Yaka"
                        },
                        {
                            "id": 902493,
                            "name": "Düğmeli Yaka"
                        },
                        {
                            "id": 1178240,
                            "name": "Choker"
                        },
                        {
                            "id": 1178764,
                            "name": "Yakasız"
                        }
                    ],
                    "categoryId": 601,
                    "required": False,
                    "varianter": False,
                    "slicer": False
                }]
        }


class GetAttributes(ChannelCommandInterface):
    """
    It is the class that enables the creation of attribute and attribute values
    without depending on category.
    """

    def get_data(self):
        data = self.objects
        return data

    def validated_data(self, data) -> object:
        return data

    def transform_data(self, data) -> object:
        return data

    def send_request(self, transformed_data) -> object:
        response = self.__mocked_request(data=transformed_data)
        return response

    def normalize_response(self, data, validated_data, transformed_data,
                           response) -> Tuple[
        List[AttributeDto], ErrorReportDto, Any]:
        """
        It is the method where the response
        from the Channel is processed and
        AttributeDto and AttributeValueDto are created.
        """
        report = self.create_report(response)

        response = response.json()
        attributes = []
        for attribute in response:
            attr = AttributeDto(remote_id=attribute["id"],
                                name=attribute["name"],
                                values=[])
            for attribute_value in attribute["attr_values"]:
                attr.values.append(
                    AttributeValueDto(remote_id=attribute_value["id"],
                                      name=attribute_value["name"]))
            attributes.append(attr)

        return attributes, report, data

    @mock_response_decorator
    def __mocked_request(self, data):
        """
        Mock a request and response for the send operation to mimic actual channel data
        """

        return [{
            "id": 123213,
            "name": "Color",
            "attr_values": [{"id": 213124,
                             "name": "Green"},
                            {"id": 21442,
                             "name": "Red"}]
        }]


class GetChannelConfSchema(ChannelCommandInterface):
    """
    Prepare channel conf schema
    """

    def get_data(self):
        return

    def validated_data(self, data) -> object:
        return data

    def transform_data(self, data) -> object:
        return data

    def send_request(self, transformed_data) -> object:
        return transformed_data

    def normalize_response(self, data, validated_data, transformed_data,
                           response) -> Tuple[dict, Any, Any]:
        schema = {
            "setting_name": ChannelConfSchemaField(
                required=True,
                data_type=ChannelConfSchemaDataTypes.bool,
                key="setting_name",
                label="setting_name"),
            "setting_name_2": ChannelConfSchemaField(
                required=True,
                data_type=ChannelConfSchemaDataTypes.text,
                key="setting_name_2",
                label="setting_name_2"),
            "reason_mapping": ChannelConfSchemaField(
                required=True,
                data_type=ChannelConfSchemaDataTypes.json,
                key="reason_mapping",
                label="reason_mapping"),
        }

        return schema, None, None



from typing import List

from omnisdk.omnitron.endpoints import (ChannelProductStockEndpoint,
                                        ChannelIntegrationActionEndpoint,
                                        ChannelBatchRequestEndpoint,
                                        ChannelExtraProductStockEndpoint,
                                        ChannelExtraProductPriceEndpoint)
from omnisdk.omnitron.models import ProductStock, Product

from channel_app.core.commands import OmnitronCommandInterface
from channel_app.core.data import BatchRequestResponseDto
from channel_app.core.utilities import split_list
from channel_app.omnitron.commands.batch_requests import ProcessBatchRequests
from channel_app.omnitron.constants import (ContentType, BatchRequestStatus,
                                            IntegrationActionStatus,
                                            FailedReasonType)


class GetUpdatedProductStocks(OmnitronCommandInterface):
    """
    Fetches updated and not sent stock objects from Omnitron

    Batch request state transition to fail or done
     :return: List[ProductStock] as output of do_action
    """
    endpoint = ChannelProductStockEndpoint
    path = "updates"
    BATCH_SIZE = 100
    content_type = ContentType.product_stock.value

    def get_data(self) -> List[ProductStock]:
        stocks = self.get_product_stocks()
        stocks = self.get_stocks_with_available(stocks)
        return stocks

    def get_product_stocks(self) -> List[ProductStock]:
        stocks = self.endpoint(
            path=self.path,
            channel_id=self.integration.channel_id
        ).list(
            params={"limit": self.BATCH_SIZE}
        )
        stocks = stocks[:self.BATCH_SIZE]
        objects_data = self.create_batch_objects(data=stocks,
                                                 content_type=self.content_type)
        self.update_batch_request(objects_data=objects_data)
        return stocks

    def get_stocks_with_available(self, stocks: List[ProductStock]):
        if not stocks:
            return []

        endpoint = ChannelIntegrationActionEndpoint(
            channel_id=self.integration.channel_id)
        stock_integration_actions = endpoint.list(
            params={
                "local_batch_id": self.integration.batch_request.local_batch_id,
                "status": IntegrationActionStatus.processing,
                "channel_id": self.integration.channel_id,
                "sort": "id"
            })
        for stock_batch in endpoint.iterator:
            stock_integration_actions.extend(stock_batch)
        stock_ia_dict = {ia.object_id: ia for ia in stock_integration_actions}
        for stock in stocks:
            stock_ia = stock_ia_dict[stock.pk]
            stock.remote_id = stock_ia.remote_id
        return stocks


class GetUpdatedProductStocksFromExtraStockList(OmnitronCommandInterface):
    """
    Fetches updated and not sent stock objects from Omnitron

    Batch request state transition to fail or done
     :return: List[ProductStock] as output of do_action
    """
    endpoint = ChannelExtraProductStockEndpoint
    path = "updates"
    BATCH_SIZE = 100
    content_type = ContentType.product_stock.value

    def get_data(self) -> List[ProductStock]:
        self.stock_list_id = self.objects
        stocks = self.get_product_stocks()
        stocks = self.get_integration_actions(stocks)
        return stocks

    def get_product_stocks(self) -> List[ProductStock]:
        endpoint = self.endpoint(path=self.path,
                                 channel_id=self.integration.channel_id)
        stocks = endpoint.list(
            params={"stock_list": self.stock_list_id,
                    "limit": self.BATCH_SIZE}
        )
        stocks = stocks[:self.BATCH_SIZE]
        objects_data = self.create_batch_objects(data=stocks,
                                                 content_type=self.content_type)
        self.update_batch_request(objects_data=objects_data)
        return stocks

    def get_integration_actions(self, stocks: List[ProductStock]):
        if not stocks:
            return []
        endpoint = ChannelIntegrationActionEndpoint(
            channel_id=self.integration.channel_id)
        stock_integration_actions = endpoint.list(
            params={
                "local_batch_id": self.integration.batch_request.local_batch_id,
                "status": IntegrationActionStatus.processing,
                "sort": "id"
            })
        for stock_batch in endpoint.iterator:
            stock_integration_actions.extend(stock_batch)
        stock_ia_dict = {ia.object_id: ia for ia in stock_integration_actions}
        for stock in stocks:
            stock_ia = stock_ia_dict[stock.pk]
            stock.remote_id = stock_ia.remote_id
        return stocks


class GetInsertedProductStocksFromExtraStockList(
    GetUpdatedProductStocksFromExtraStockList):
    """
    Fetches inserted stock data from Omnitron.

    Batch request state transition to fail or done
     :return: List[ProductStock] as output of do_action
    """
    path = "inserts"

    def get_integration_actions(self, stocks: List[ProductStock]):
        if not stocks:
            return []

        endpoint = ChannelIntegrationActionEndpoint(
            channel_id=self.integration.channel_id)
        product_ids = [str(stock.product) for stock in stocks]
        product_ias = endpoint.list(
            params={"object_id__in": ",".join(product_ids),
                    "content_type_name": ContentType.product.value,
                    "channel_id": self.integration.channel_id,
                    "sort": "id"
                    })

        for product_batch in endpoint.iterator:
            product_ias.extend(product_batch)
        product_integrations_by_id = {ia.object_id: ia for ia in product_ias}

        for stock in stocks:
            if stock.product in product_integrations_by_id:
                product_ia = product_integrations_by_id[stock.product]
                stock.remote_id = product_ia.remote_id
            else:
                stock.failed_reason_type = FailedReasonType.channel_app.value
                self.failed_object_list.append(
                    (stock, ContentType.product_stock.value,
                     "Product has not been sent"))
        return stocks


class GetInsertedProductStocks(GetUpdatedProductStocks):
    """
    Fetches inserted stock data from Omnitron.

    Batch request state transition to fail or done

     :return: List[ProductStock] as output of do_action
    """
    path = "inserts"

    def get_stocks_with_available(self, stocks: List[ProductStock]):
        if not stocks:
            return []

        endpoint = ChannelIntegrationActionEndpoint(
            channel_id=self.integration.channel_id)
        product_ids = [str(stock.product) for stock in stocks]
        product_ias = endpoint.list(
            params={"object_id__in": ",".join(product_ids),
                    "content_type_name": ContentType.product.value,
                    "status": IntegrationActionStatus.success,
                    "channel_id": self.integration.channel_id,
                    "sort": "id"
                    })
        for product_batch in endpoint.iterator:
            product_ias.extend(product_batch)
        product_integrations_by_id = {ia.object_id: ia for ia in product_ias}

        for stock in stocks:
            if stock.product in product_integrations_by_id:
                product_ia = product_integrations_by_id[stock.product]
                stock.remote_id = product_ia.remote_id
            else:
                stock.failed_reason_type = FailedReasonType.channel_app.value
                self.failed_object_list.append(
                    (stock, ContentType.product_stock.value,
                     "Product has not been sent"))
        return stocks


class GetProductPricesFromProductStocks(OmnitronCommandInterface):
    endpoint = ChannelExtraProductPriceEndpoint
    CHUNK_SIZE = 50
    content_type = ContentType.product_price.value

    def get_data(self) -> List[ProductStock]:
        product_stocks = self.objects
        self.get_product_price(product_stocks)
        return product_stocks

    def normalize_response(self, data, response) -> List[object]:
        object_list = []
        failed_stocks = [failed_product_stocks[0] for failed_product_stocks in
                         self.failed_object_list]
        product_stock_object_list = self.create_batch_objects(
            data=failed_stocks,
            content_type=ContentType.product_stock.value)
        object_list.extend(product_stock_object_list)

        self.create_integration_actions(data, object_list)
        self.update_batch_request(object_list)
        return data

    def create_integration_actions(self, data, object_list):
        commit_product_prices = [product_stocks.productprice for product_stocks
                                 in data
                                 if not getattr(product_stocks,
                                                "failed_reason_type",
                                                None)]
        product_price_object_list = self.create_batch_objects(
            data=commit_product_prices,
            content_type=ContentType.product_price.value)
        object_list.extend(product_price_object_list)

    def get_product_price(self, product_stocks: List[ProductStock]) -> List[
        ProductStock]:
        if not product_stocks:
            empty_list: List[ProductStock] = []
            return empty_list

        endpoint = self.endpoint(channel_id=self.integration.channel_id)
        product_ids = []
        for ps in product_stocks:
            if not getattr(ps, "failed_reason_type", None):
                if isinstance(ps.product, Product):
                    product_ids.append(str(ps.product.pk))
                else:
                    product_ids.append(str(ps.product))

        prices = []
        for chunk in split_list(product_ids, self.CHUNK_SIZE):
            price_batch = self.get_prices(chunk, endpoint)
            if not price_batch:
                break
            prices.extend(price_batch)

        product_prices = {s.product: s for s in prices}

        for index, product_stock in enumerate(product_stocks):
            if getattr(product_stock, "failed_reason_type", None):
                continue
            try:
                if isinstance(product_stock.product, Product):
                    product_stock.productprice = product_prices[
                        product_stock.product.pk]
                else:
                    product_stock.productprice = product_prices[
                        product_stock.product]
            except KeyError:
                product_stock.failed_reason_type = FailedReasonType.channel_app.value
                self.failed_object_list.append(
                    (product_stock, ContentType.product_stock.value,
                     "PriceNotFound"))
                continue

        return product_stocks

    def get_prices(self, chunk, endpoint):
        price_list = getattr(self, "param_price_list",
                             self.integration.catalog.price_list)
        price_batch = endpoint.list(params={"product__pk__in": ",".join(chunk),
                                            "price_list": price_list,
                                            "limit": len(chunk)})
        return price_batch


class ProcessStockBatchRequests(OmnitronCommandInterface, ProcessBatchRequests):
    """
    Processes response of the channel for the stock update/insert task.
    It updates the stock and related model integration action data according to the
    channel response.

    As input;

    objects : List[BatchRequestResponseDto]. Represents channel response for the
    stock items that are in the BatchRequest.

    batch_request: If related batch request is not supplied, Omnitron will not
    be notified of the results.
    omnitron_integration.batch_request = batch_request

    Batch Request state transition to fail or done

     :return: None
    """
    endpoint = ChannelBatchRequestEndpoint
    content_type = ContentType.product_stock.value
    CHUNK_SIZE = 50
    BATCH_SIZE = 100

    def get_data(self):
        return self.objects

    def validated_data(self, data: List[BatchRequestResponseDto]):
        for item in data:
            assert isinstance(item, BatchRequestResponseDto)
        return data

    def send(self, validated_data):
        self.process_item(validated_data)

    @property
    def update_state(self, *args, **kwargs) -> BatchRequestStatus:
        return BatchRequestStatus.done

    def check_run(self, is_ok, formatted_data):
        if not is_ok and self.is_batch_request:
            self.integration.batch_request.objects = None
            self.batch_service(self.integration.channel_id).to_fail(
                self.integration.batch_request)
        return False

    def get_channel_items_by_reference_object_ids(self, channel_response,
                                                  model_items_by_content,
                                                  integration_actions):
        product_ids = [str(item) for item in
                       model_items_by_content["productstock"]]

        model_items_by_content_product = self.get_products(product_ids)

        channel_items_by_product_id = {}
        for product_id, product in model_items_by_content_product.items():
            sku = self.get_barcode(obj=product)
            for channel_item in channel_response:
                # TODO: comment
                if channel_item.sku != sku:
                    continue
                remote_item = channel_item
                channel_items_by_product_id[product_id] = remote_item
                break
        return channel_items_by_product_id



import datetime
from random import random
from typing import Tuple, Any, List

from omnisdk.omnitron.models import Order, BatchRequest

from channel_app.core.commands import ChannelCommandInterface
from channel_app.core.data import (BatchRequestResponseDto, 
                                   CancellationRequestDto, 
                                   ChannelCancellationRequestDto, 
                                   ErrorReportDto, 
                                   ChannelCreateOrderDto, AddressDto, 
                                   CustomerDto, ChannelOrderDto, OrderItemDto, 
                                   OrderBatchRequestResponseDto,
                                   CancelOrderDto, ChannelUpdateOrderItemDto)
from channel_app.omnitron.constants import ResponseStatus


class GetOrders(ChannelCommandInterface):
    def get_data(self):

        data = self.objects
        return data

    def validated_data(self, data) -> object:
        return data

    def transform_data(self, data) -> object:
        return data

    def send_request(self, transformed_data) -> object:
        response = self.__mocked_request(data=transformed_data)
        return response

    def normalize_response(self, data, validated_data, transformed_data,
                           response) -> Tuple[ChannelCreateOrderDto,
                                              ErrorReportDto, Any]:
        """
        Convert channel order to the format OmnitronIntegration
        requires
        """
        for response_order_data in response:
            order_data = response_order_data["order"]
            customer_data = order_data.pop("customer")
            shipping_address_data = order_data.pop("shipping_address")
            billing_address_data = order_data.pop("billing_address")

            order_items_data = response_order_data["order_items"]

            shipping_address = AddressDto(**shipping_address_data)

            billing_address = AddressDto(**billing_address_data)

            customer = CustomerDto(**customer_data)

            channel_order_data = {
                "billing_address": billing_address,
                "shipping_address": shipping_address,
                "customer": customer,
                "cargo_company": order_data["cargo_company"]
            }
            channel_order_data.update(order_data)
            channel_order_dto = ChannelOrderDto(**channel_order_data)

            order_items_dto = []
            for order_item_data in order_items_data:
                order_item = OrderItemDto(**order_item_data)
                order_items_dto.append(order_item)

            order_items_dto: List[OrderItemDto]
            channel_create_order = ChannelCreateOrderDto(
                order=channel_order_dto,
                order_item=order_items_dto)

            report = self.create_report(response_order_data)
            yield channel_create_order, report, None

    def __mocked_request(self, data):
        number = str(int(random() * 1000000))
        return [{
            "order": {
                "status": "400",
                "remote_id": number,
                "number": number,
                "channel": "1",
                "currency": "try",
                "amount": "17",
                "shipping_amount": "0.0",
                "shipping_tax_rate": "18",
                "extra_field": {},
                "created_at": str(datetime.datetime.now()),
                "customer": {
                    "email": "dummy@dummy.com",
                    "phone_number": None,
                    "first_name": "Dummy",
                    "last_name": "Dummy",
                    "channel_code": "1212",
                    "extra_field": None,
                    "is_active": True
                },
                "shipping_address": {
                    "email": "dummy@dummy.com",
                    "phone_number": "05540000000",
                    "first_name": "dummy",
                    "last_name": "dummy",
                    "country": "Türkiye",
                    "city": "İstanbul",
                    "line": "dummy 3 dummy cd"
                },
                "billing_address": {
                    "email": "dummy@dummy.com",
                    "phone_number": "05540000000",
                    "first_name": "dummy",
                    "last_name": "dummy",
                    "country": "Türkiye",
                    "city": "İstanbul",
                    "line": "dummy 3 dummy cd"
                },
                "cargo_company": "aras"
            },
            "order_items": [
                {
                    "remote_id": "1234567816",
                    "product": "1234",
                    "price_currency": "try",
                    "price": "17.0",
                    "tax_rate": "18",
                    "retail_price": "20.0",
                    "extra_field": {"tracking_number": "1231"},
                    "status": "400"
                }
            ]
        }]


class SendUpdatedOrders(ChannelCommandInterface):
    param_sync = True

    def get_data(self) -> List[Order]:
        data = self.objects
        return data

    def validated_data(self, data) -> object:
        return data

    def transform_data(self, data) -> object:
        return data

    def send_request(self, transformed_data) -> object:
        response = self.__mocked_request(data=transformed_data)
        return response

    def normalize_response(self, data, validated_data, transformed_data,
                           response) -> Tuple[
            List[OrderBatchRequestResponseDto], List[ErrorReportDto], Any]:
        response_data = []
        for row in response:
            obj = OrderBatchRequestResponseDto(
                status=row["status"],
                remote_id=row["remote_id"],
                number=row["number"],
                message=row["message"])
            response_data.append(obj)

        report = self.create_report(response)
        return response_data, report, data

    def __mocked_request(self, data):
        """
        Mock a request and response for the send operation to mimic actual
        channel data

        :return:

        [{
            "status": "SUCCESS",
            "remote_id": "123a1",
            "number": "1234567",
            "message": ""
         },]
        """
        response_data = []
        prefix = "order"
        for index, item in enumerate(data):
            if random() < 0.8:
                response_item = {
                    'number': item.number,
                    'message': "",
                    'remote_id': "{}_{}".format(prefix, index),
                    'status': ResponseStatus.success
                }
            else:
                response_item = {
                    "status": ResponseStatus.fail,
                    "remote_id": None,
                    "number": item.number,
                    "message": "exception message"
                }
            response_data.append(response_item)
        return response_data


class CheckOrders(ChannelCommandInterface):
    def get_data(self) -> BatchRequest:
        batch_request = self.objects
        return batch_request

    def validated_data(self, data) -> object:
        return data

    def transform_data(self, data) -> object:
        return data

    def send_request(self, transformed_data: BatchRequest) -> object:
        """
            Sends a post request to the channel client to insert the products
        :param transformed_data:
        :return:
            [{ remote_data},]
        """

        response = self.__mocked_request(
            data=self.integration._sent_data[transformed_data.remote_batch_id],
            remote_batch_id=transformed_data.remote_batch_id)
        return response

    def normalize_response(self, data, validated_data, transformed_data,
                           response) -> Tuple[
            List[OrderBatchRequestResponseDto], List[ErrorReportDto], Any]:
        response_data = []
        for row in response:
            obj = OrderBatchRequestResponseDto(
                status=row["status"],
                remote_id=row["remote_id"],
                number=row["number"],
                message=row["message"])
            response_data.append(obj)

        report = self.create_report(response)
        return response_data, report, data

    def __mocked_request(self, data, remote_batch_id):
        """
        Mock a request and response for the send operation to mimic actual
        channel data

        :return:

        [{
            "status": "SUCCESS",
            "remote_id": "123a1",
            "number": "1234567",
            "message": ""
         },]
        """
        response_data = []
        prefix = remote_batch_id[-8:]
        for index, item in enumerate(data):
            if random() < 0.8:
                response_item = {
                    'number': item.number,
                    'message': "",
                    'remote_id': "{}_{}".format(prefix, index),
                    'status': ResponseStatus.success
                }
            else:
                response_item = {
                    "status": ResponseStatus.fail,
                    "remote_id": None,
                    "number": item.number,
                    "message": "exception message"
                }
            response_data.append(response_item)
        return response_data


class GetCancelledOrders(ChannelCommandInterface):
    def get_data(self) -> None:
        return self.objects

    def validated_data(self, data) -> object:
        return data

    def transform_data(self, data) -> object:
        return data

    def send_request(self, transformed_data: BatchRequest) -> object:
        """
            Sends a post request to the channel client to insert the products
        :param transformed_data:
        :return:
            [{ remote_data},]
        """

        response = self.__mocked_request(
            data=self.integration._sent_data[transformed_data],
            remote_batch_id=transformed_data)
        return response

    def normalize_response(self, data, validated_data, transformed_data,
                           response) -> Tuple[List[CancelOrderDto],
                                              ErrorReportDto, Any]:
        report = self.create_report(response)

        for row in response:
            obj = CancelOrderDto(
                order=row["order"],
                cancel_items=row["cancel_items"],
                reasons=row["reasons"]
            )

            yield obj, report, data

    def __mocked_request(self, data, remote_batch_id):
        """
        Mock a request and response for the send operation to mimic actual
        channel data

        :return:

        [{
            "order": "orderNumber123",
            "cancel_items": ["remote_item_1", "remote_item_2"],
            "reasons": {"remote_item_1": "remote_reason_code",
                        "remote_item_2": "remote_reason_code"}
         },]
        """
        response_data = []
        for index, item in enumerate(data):
            response_item = {
                'order': item.number,
                'cancel_items': item.cancel_items,
                'reasons': item.reasons,
            }
            response_data.append(response_item)
        return response_data


class GetUpdatedOrderItems(ChannelCommandInterface):
    def get_data(self):
        data = self.objects
        return data

    def validated_data(self, data) -> object:
        return data

    def transform_data(self, data) -> object:
        return data

    def send_request(self, transformed_data) -> object:
        response = self.__mocked_request(data=transformed_data)
        return response

    def normalize_response(self, data, validated_data, transformed_data,
                           response) -> Tuple[ChannelUpdateOrderItemDto,
                                              ErrorReportDto, Any]:
        """
        Convert ChannelUpdateOrderItemDto to the format OmnitronIntegration
        """
        for order_item_data in response:
            report = self.create_report(order_item_data)
            channel_update_order_item = ChannelUpdateOrderItemDto(
                **order_item_data)

            yield channel_update_order_item, report, None

    def __mocked_request(self, data):
        return [{
            "remote_id": "XYAD123213",
            "order_remote_id": "1234",
            "order_number": "1234",
            "status": "550",
            "invoice_number": "1234",
            "invoice_date": None,
            "tracking_number": "400",
            "extra_field": {"tracking_number": "1231"},
        }]


class GetCancellationRequests(ChannelCommandInterface):
    def get_data(self):
        data = self.objects
        return data

    def validated_data(self, data) -> object:
        return data

    def transform_data(self, data) -> object:
        return data

    def send_request(self, transformed_data) -> object:
        response = self.__mocked_request(data=transformed_data)
        return response

    def normalize_response(
            self, data, validated_data, transformed_data,
            response) -> Tuple[CancellationRequestDto, ErrorReportDto, Any]:
        report = self.create_report(response)
        for row in response:
            obj = CancellationRequestDto(
                order_item=row["order_item"],
                reason=row["reason"],
                cancellation_type=row["cancellation_type"],
                remote_id=row["remote_id"]
            )

            yield obj, report, data

    def __mocked_request(self, data):
        """
        Mock a request and response for the send operation to mimic actual
        channel data

        :return:

        [{
            "order_item": "remote_item_1",
            "reason": "remote_reason_code",
            "cancellation_type": "remote_cancellation_type"
         },]
        """
        response_data = [
            {
                'order_item': "1234567816",
                'reason': "iptal",
                'cancellation_type': "refund",
                'remote_id': "1234567816_1"
            }
        ]
        return response_data


class UpdateCancellationRequest(ChannelCommandInterface):
    def get_data(self) -> ChannelCancellationRequestDto:
        isinstance(self.objects, ChannelCancellationRequestDto)
        data = self.objects
        return data

    def validated_data(self, data) -> object:
        return data

    def transform_data(self, data) -> object:
        return data

    def send_request(self, transformed_data) -> object:
        response = self.__mocked_request(data=transformed_data)
        return response

    def normalize_response(
            self, data, validated_data, transformed_data,
            response) -> Tuple[BatchRequestResponseDto, ErrorReportDto, Any]:

        obj = BatchRequestResponseDto(
            status=response["status"],
            remote_id=response["remote_id"],
            sku=response["sku"],
            message=response["message"])
        
        report = self.create_report(response)
        return obj, report, data

    def __mocked_request(self, data):
        """
        Mock a request and response for the send operation to mimic actual
        channel data

        :return:
            
            {
                "status": "SUCCESS",
                "remote_id": "123a1",
                "sku": "1234567",
                "message": ""
            }
        """
        return {
            "status": ResponseStatus.success,
            "remote_id": data.remote_id,
            "sku": data.remote_id,
            "message": data.status
        }
        



import asyncio
import uuid
from functools import lru_cache

import requests
from omnisdk.omnitron.endpoints import ChannelAttributeSetEndpoint, \
    ChannelAttributeEndpoint, \
    ChannelAttributeConfigEndpoint, ChannelAttributeValueEndpoint, \
    ChannelAttributeSetConfigEndpoint, ChannelAttributeValueConfigEndpoint, \
    ChannelAttributeSchemaEndpoint
from omnisdk.omnitron.endpoints import ChannelCategoryTreeEndpoint, \
    ChannelCategoryNodeEndpoint, \
    ContentTypeEndpoint, ChannelIntegrationActionEndpoint, ChannelEndpoint
from omnisdk.omnitron.models import CategoryNode, IntegrationAction, \
    CategoryTree, Channel, \
    ChannelAttributeSetConfig, ChannelAttributeValueConfig, \
    ChannelAttributeSchema
from omnisdk.omnitron.models import ChannelAttributeSet, ChannelAttribute, \
    ChannelAttributeConfig, \
    ChannelAttributeValue
from requests import HTTPError

from channel_app.core.commands import OmnitronCommandInterface
from channel_app.core.data import CategoryTreeDto
from channel_app.core.utilities import is_updated, split_list
from channel_app.omnitron.constants import ContentType


class CreateOrUpdateCategoryTreeAndNodes(OmnitronCommandInterface):
    """
    Using the channel category tree data (including all nodes)
    Create/Update category tree and nodes on the Omnitron side.
    """
    endpoint = ChannelCategoryTreeEndpoint

    def get_data(self) -> CategoryTreeDto:
        return self.objects

    def send(self, validated_data):
        """
        Iterative tree traversal implementation with stack by processing parents first
        children later.
        :return:
        """
        # TODO what should we do once a leaf node becomes a parent node?
        #  How do we update this node? What happens to the products that are
        #  assigned to this category?

        tree = validated_data
        channel_endpoint = ChannelEndpoint()

        if not self.integration.channel.category_tree:
            ct = CategoryTree()
            ct.name = "Sales Channel Tree {}".format(str(uuid.uuid4()))
            node = self.endpoint(channel_id=self.integration.channel_id).create(
                item=ct)
            channel = Channel(category_tree=node.pk)
            channel_endpoint.update(id=self.integration.channel_id,
                                    item=channel)
            self.integration.channel_object = None
        stack = []
        current = tree.root
        node_endpoint = ChannelCategoryNodeEndpoint(
            channel_id=self.integration.channel_id)
        content_type = \
            ContentTypeEndpoint().list(params={"model": "categorynode"})[0]
        integration_action_endpoint = ChannelIntegrationActionEndpoint(
            channel_id=self.integration.channel_id)
        while True:
            if current:
                integration_action = self.get_integration_action(
                    content_type_model=content_type.model,
                    remote_id=current.remote_id)
                if integration_action:  # update node
                    node_object = CategoryNode()
                    node_object.name = current.name
                    node_endpoint.update(id=integration_action.object_id,
                                         item=node_object)
                    current.omnitron_id = integration_action.object_id
                elif not current.remote_id:
                    # if block to skip node creation for root node
                    pass

                else:  # create node
                    self.create_node(content_type, current,
                                     integration_action_endpoint,
                                     node_endpoint)
                stack.extend(current.children)
                current = None
            elif stack:
                current = stack.pop()
            else:
                break
        return []

    def create_node(self, content_type, current, integration_action_endpoint,
                    node_endpoint) -> (
            CategoryNode, IntegrationAction):

        parent_remote_id = current.parent and current.parent.remote_id

        if parent_remote_id:
            node_object_parent = self.get_integration_action(
                content_type_model=content_type.model,
                remote_id=parent_remote_id)
            node_object_parent_id = node_object_parent.object_id
        else:
            root_node = self.root_node()
            node_object_parent_id = root_node["pk"]

        node_object_name = current.name

        node_object = CategoryNode()
        node_object.node = node_object_parent_id
        node_object.name = node_object_name
        node = None
        try:
            node = node_endpoint.create(item=node_object)
        except requests.HTTPError as e:
            if e.response.status_code // 100 != 4:
                raise e

            parent_node_detailed = self.get_node(node_id=node_object_parent_id,
                                                 is_detailed=True)
            for child_node in parent_node_detailed.children:
                if child_node["name"] == node_object_name:
                    node = self.get_node(node_id=child_node["pk"])
                    break

            if not node:
                raise e

        integration_action = IntegrationAction(
            channel_id=self.integration.channel_id,
            content_type_id=content_type.id,
            version_date=node.modified_date,
            object_id=node.pk,
            remote_id=current.remote_id)

        integration_action = integration_action_endpoint.create(
            item=integration_action)
        current.omnitron_id = node.pk
        return node, integration_action

    def get_integration_action(self, content_type_model, remote_id):
        if not remote_id:
            return None
        integration_action_endpoint = ChannelIntegrationActionEndpoint(
            channel_id=self.integration.channel_id)

        integration_action = integration_action_endpoint.list(params={
            "channel_id": self.integration.channel_id,
            "content_type_name": content_type_model,
            "remote_id__exact": remote_id
        })
        if integration_action:
            return integration_action[0]
        else:
            return None

    def get_node(self, node_id, is_detailed=False) -> CategoryNode:
        path = None
        if is_detailed:
            path = "detailed"
        node_endpoint = ChannelCategoryNodeEndpoint(
            channel_id=self.integration.channel_id,
            path=path)
        detailed_node = node_endpoint.retrieve(id=node_id)
        return detailed_node

    def root_node(self):
        category_tree_endpoint = ChannelCategoryTreeEndpoint(
            channel_id=self.integration.channel_id)
        return category_tree_endpoint.retrieve(
            id=self.integration.channel.category_tree).category_root

    def check_run(self, is_ok, formatted_data):
        return True


class CreateOrUpdateCategoryAttributes(OmnitronCommandInterface):
    """
    Create Attribute related entries on the Omnitron using the
    attribute data created from a CategoryDto object.
    """
    integration_action_endpoint = ChannelIntegrationActionEndpoint

    def get_data(self):
        return self.objects

    def send(self, validated_data):
        integration_action, channel_category = validated_data
        if not channel_category:
            self.update_category_node_version_date(integration_action)
            return [integration_action]
        attribute_set_name = self.get_attribute_set_name(channel_category)
        attribute_set = self.integration.do_action(
            key="create_or_update_channel_attribute_set",
            objects={"name": attribute_set_name,
                     "remote_id": integration_action.remote_id})[0]
        self.integration.do_action(
            key="get_or_create_channel_attribute_set_config",
            objects={"attribute_set": attribute_set.id,
                     "object_id": integration_action.object_id,
                     "content_type": ContentType.category_node.value})

        for channel_attribute in channel_category.attributes:
            attribute = self.integration.do_action(
                key="create_or_update_channel_attribute",
                objects={"name": channel_attribute.name,
                         "remote_id": channel_attribute.remote_id})[0]
            schema = self.integration.do_action(
                key="get_or_create_channel_attribute_schema",
                objects={"name": f"{attribute.name} Schema"}
            )
            self.integration.do_action(
                key="create_or_update_channel_attribute_config",
                objects={"attribute": attribute.pk,
                         "attribute_set": attribute_set.id,
                         "attribute_remote_id": channel_attribute.remote_id,
                         "is_required": channel_attribute.required,
                         "is_custom": channel_attribute.allow_custom_value,
                         "is_variant": channel_attribute.variant,
                         })
            for channel_attribute_value in channel_attribute.values:
                self.create_attribute_value_and_config(
                    attribute=attribute, attribute_set=attribute_set,
                    channel_attribute_value=channel_attribute_value)

        self.update_category_node_version_date(integration_action)
        return [attribute]

    def update_category_node_version_date(self, integration_action):
        category_node_endpoint = ChannelCategoryNodeEndpoint(
            channel_id=self.integration.channel_id)
        category_node = category_node_endpoint.update(
            id=integration_action.object_id,
            item=CategoryNode())
        self.integration_action_endpoint(
            channel_id=self.integration.channel_id).update(
            id=integration_action.pk,
            item=IntegrationAction(version_date=category_node.modified_date))

    def create_attribute_value_and_config(self, attribute, attribute_set,
                                          channel_attribute_value):
        attribute_value = self.integration.do_action(
            key="create_or_update_channel_attribute_value",
            objects={
                "attribute": attribute.pk,
                "label": channel_attribute_value.name,
                "value": channel_attribute_value.remote_id,
                "remote_id": channel_attribute_value.remote_id})[0]
        attribute_value_config = self.integration.do_action(
            key="get_or_create_channel_attribute_value_config",
            objects={
                "attribute_value": attribute_value.pk,
                "attribute_set": attribute_set.id})[0]

    def get_attribute_set_name(self, channel_category):
        # TODO create attribute set name as a breadcrumb
        return channel_category.name


class AsyncCreateOrUpdateCategoryAttributes(CreateOrUpdateCategoryAttributes):
    """
    Async process
    Create Attribute related entries on the Omnitron using the
    attribute data created from a CategoryDto object.
    """

    async def run_async(self):
        integration_action, channel_category = self.get_data()
        attribute_set_name = self.get_attribute_set_name(channel_category)
        attribute_set = self.integration.do_action(
            key="create_or_update_channel_attribute_set",
            objects={"name": attribute_set_name,
                     "remote_id": integration_action.remote_id})
        self.integration.do_action(
            key="get_or_create_channel_attribute_set_config",
            objects={"attribute_set": attribute_set.remote_id,
                     "object_id": integration_action.object_id,
                     "content_type": ContentType.category_node.value})

        for channel_attribute in channel_category.attributes:
            attribute = self.integration.do_action(
                key="create_or_update_channel_attribute",
                objects={"name": channel_attribute.name,
                         "remote_id": channel_attribute.remote_id})
            self.integration.do_action(
                key="create_or_update_channel_attribute_config",
                objects={"attribute": attribute.pk,
                         "attribute_set": attribute_set.remote_id,
                         "attribute_remote_id": channel_attribute.remote_id,
                         "is_required": channel_attribute.required,
                         "is_custom": channel_attribute.allow_custom_value,
                         "is_variant": channel_attribute.variant,
                         })

            await asyncio.gather(
                *[self.get_or_create_attribute_value_and_config(attribute,
                                                                attribute_set,
                                                                channel_attribute_value)
                  for
                  channel_attribute_value in channel_attribute.values])

        category_node_endpoint = ChannelCategoryNodeEndpoint(
            channel_id=self.integration.channel_id)
        category_node = category_node_endpoint.update(
            id=integration_action.object_id,
            item=CategoryNode())
        self.integration_action_endpoint(
            channel_id=self.integration.channel_id).update(
            id=integration_action.pk,
            item=IntegrationAction(version_date=category_node.modified_date))

    async def get_or_create_attribute_value_and_config(self, attribute,
                                                       attribute_set,
                                                       channel_attribute_value):
        attribute_value = self.integration.do_action(
            key="create_or_update_channel_attribute_value",
            objects={
                "attribute": attribute.pk,
                "label": channel_attribute_value.name,
                "value": channel_attribute_value.remote_id,
                "remote_id": channel_attribute_value.remote_id})
        attribute_value_config = self.integration.do_action(
            key="get_or_create_channel_attribute_value_config",
            objects={
                "attribute_value": attribute_value.pk,
                "attribute_set": attribute_set.remote_id})
        await asyncio.sleep(0)


class GetCategoryIds(OmnitronCommandInterface):
    """
    Get Category ids from Omnitron sorted by least recently updated category first
    """
    endpoint = ChannelCategoryNodeEndpoint
    CHUNK_SIZE = 50

    def get_data(self):
        return

    def send(self, validated_data) -> list:
        if not self.integration.channel.category_tree:
            raise Exception("No category tree, no attributes :/")
        integration_endpoint = ChannelIntegrationActionEndpoint(
            channel_id=self.integration.channel_id)

        integration_actions = integration_endpoint.list(params={
            "channel_id": self.integration.channel_id,
            "content_type_name": ContentType.category_node.value,
            "sort": "version_date",
            "limit": 200  # if we want to process oldest five nodes at a time
        })
        return integration_actions


class CreateOrUpdateChannelAttributeSet(OmnitronCommandInterface):
    """
    Create AttributeSet object on Omnitron to store the attributes
    """
    endpoint = ChannelAttributeSetEndpoint

    def get_data(self) -> ChannelAttributeSet:
        name = self.objects["name"]
        remote_id = self.objects["remote_id"]
        channel = self.integration.channel_id
        return ChannelAttributeSet(name=name, remote_id=remote_id,
                                   channel=channel)

    def send(self, validated_data) -> list:
        data = validated_data
        endpoint = self.endpoint(channel_id=self.integration.channel_id)

        attribute_sets = endpoint.list(
            params={"remote_id__exact": data.remote_id,
                    "channel": self.integration.channel_id})
        # TODO add channel filter too on the next version

        if len(attribute_sets) > 0:
            attribute_set = attribute_sets[0]
            if self.is_updated(attribute_set, data):
                self.endpoint(channel_id=self.integration.channel_id).update(
                    id=attribute_set.id, item=data)
            return [attribute_sets[0]]

        attribute_set = self.endpoint(
            channel_id=self.integration.channel_id).create(item=data)
        return [attribute_set]

    def is_updated(self, attribute_set, data):
        if data.channel != attribute_set.channel:
            return True
        if data.name != attribute_set.name:
            return True
        return False


class GetOrCreateChannelAttributeSetConfig(OmnitronCommandInterface):
    """
    Create an AttributeSetConfig entry for AttributeSet denoting which object this set belongs to.
    For example an attribute set usually defines the attributes a category needs. So
    we create an AttributeSetConfig entry with category Omnitron id, CategoryNode content type and
    AttributeSet id.
    """

    endpoint = ChannelAttributeSetConfigEndpoint

    def get_data(self) -> ChannelAttributeSetConfig:
        attribute_set = self.objects["attribute_set"]
        object_id = self.objects["object_id"]
        content_type = self.objects["content_type"]
        if not isinstance(attribute_set, int):
            raise AssertionError
        if not isinstance(object_id, (int, type(None))):
            raise AssertionError
        if not isinstance(content_type, str):
            raise AssertionError
        return ChannelAttributeSetConfig(attribute_set=attribute_set,
                                         object_id=object_id,
                                         content_type=content_type)

    def send(self, validated_data) -> list:
        data = validated_data
        endpoint = self.endpoint(channel_id=self.integration.channel_id)
        attribute_set_configs = endpoint.list(
            params={"attribute_set": data.attribute_set,
                    "object_id": data.object_id,
                    "content_type__model__exact": data.content_type})
        if len(attribute_set_configs) > 0:
            return attribute_set_configs

        attribute_set_config = endpoint.create(item=data)
        return [attribute_set_config]


class CreateOrUpdateChannelAttribute(OmnitronCommandInterface):
    endpoint = ChannelAttributeEndpoint

    # TODO Omnitron does not support 2 different attributes with same name on same channel,
    #  If a marketplace uses more than one same-name attributes for different purposes,
    #  in Omnitron they map to a single attribute. Only one of those attributes will be recognized
    #  and an IntegrationAction entry will be created.

    def get_data(self):
        name = self.objects["name"]
        remote_id = self.objects["remote_id"]
        if not isinstance(name, str):
            raise AssertionError
        if not isinstance(remote_id, (str, int, type(None))):
            raise AssertionError

        return name, remote_id

    def send(self, validated_data) -> list:
        name, remote_id = validated_data

        integration_endpoint = ChannelIntegrationActionEndpoint(
            channel_id=self.integration.channel_id)

        attribute = self.create_or_update_channel_attribute(
            integration_endpoint, name, remote_id)
        return [attribute]

    def create_or_update_channel_attribute(self, integration_endpoint, name,
                                           remote_id):

        integration_action = integration_endpoint.list(params={
            "content_type_name": "marketplaceattribute",
            "channel_id": self.integration.channel_id,
            "remote_id__exact": remote_id
        })
        if integration_action:
            integration_action = integration_action[0]
            attribute = self.endpoint(
                channel_id=self.integration.channel_id).retrieve(
                id=integration_action.object_id)
            if attribute.name != name:
                attribute = self.endpoint(
                    channel_id=self.integration.channel_id).update(
                    id=integration_action.object_id, item=ChannelAttribute(
                        channel=self.integration.channel_id,
                        name=name
                    ))
            return attribute
        else:
            try:
                attribute = self.endpoint(
                    channel_id=self.integration.channel_id).create(
                    item=ChannelAttribute(channel=self.integration.channel_id,
                                          name=name))
            except HTTPError as e:
                if e.response.status_code // 100 != 4:
                    raise e
                params = {
                    "name__exact": name, "channel": self.integration.channel_id
                }
                attributes = self.endpoint(
                    channel_id=self.integration.channel_id).list(params=params)
                if len(attributes) == 0:
                    raise Exception(
                        "Creation returned bad request but nothing exists!")
                return attributes[0]
            content_type = ContentTypeEndpoint().list(
                params={"model": "marketplaceattribute"})[0]
            integration_endpoint.create(item=IntegrationAction(
                channel_id=self.integration.channel_id,
                content_type_id=content_type.id,
                remote_id=remote_id,
                object_id=attribute.pk,
                version_date=attribute.modified_date,
            ))
        return attribute


class GetChannelAttributeSets(OmnitronCommandInterface):
    """
      Get all attribute set  in omnitron
      """

    def get_data(self):
        return

    def send(self, validated_data) -> object:
        endpoint = ChannelAttributeSetEndpoint(
            channel_id=self.integration.channel_id)
        attribute_sets = endpoint.list(
            params={"channel": self.integration.channel_id,
                    "sort": "id"})
        for attribute_set_batch in endpoint.iterator:
            attribute_sets.extend(attribute_set_batch)
        return attribute_sets


class GetChannelAttributeSetConfigs(OmnitronCommandInterface):
    """
    Get all attribute set config in omnitron
    """
    CHUNK_SIZE = 500

    def get_data(self):
        return

    def send(self, validated_data) -> object:
        endpoint = ChannelAttributeSetConfigEndpoint(
            channel_id=self.integration.channel_id)
        attribute_set_configs = endpoint.list(params={
            "content_type__model": ContentType.category_node.value,
            "limit": self.CHUNK_SIZE,
            "sort": "object_id"})
        for attribute_set_config_batch in endpoint.iterator:
            attribute_set_configs.extend(attribute_set_config_batch)
        attribute_set_configs_by_category_node_pk = {}
        for attribute_set_config in attribute_set_configs:
            attribute_set_configs_by_category_node_pk[
                attribute_set_config.object_id] = attribute_set_config
        return [attribute_set_configs_by_category_node_pk]


class CreateOrUpdateChannelAttributeConfig(OmnitronCommandInterface):
    """
    Attach an attribute with an attribute set
    """

    endpoint = ChannelAttributeConfigEndpoint

    def get_data(self) -> dict:
        attribute_set = self.objects["attribute_set"]
        attribute = self.objects["attribute"]
        attribute_remote_id = self.objects.get("attribute_remote_id", None)
        is_required = self.objects.get("is_required", False)
        is_variant = self.objects.get("is_variant", False)
        is_custom = self.objects.get("is_custom", False)
        is_image_attribute = self.objects.get("is_image_attribute", False)
        is_meta = self.objects.get("is_meta", False)

        if not isinstance(attribute_set, int):
            raise AssertionError
        if not isinstance(attribute, int):
            raise AssertionError
        if not isinstance(attribute_remote_id, (str, type(None))):
            raise AssertionError
        if not isinstance(is_required, bool):
            raise AssertionError
        if not isinstance(is_variant, bool):
            raise AssertionError
        if not isinstance(is_custom, bool):
            raise AssertionError
        if not isinstance(is_image_attribute, bool):
            raise AssertionError
        if not isinstance(is_meta, bool):
            raise AssertionError
        return {
            "attribute_set": attribute_set,
            "attribute": attribute,
            "attribute_remote_id": attribute_remote_id,
            "is_required": is_required,
            "is_variant": is_variant,
            "is_custom": is_custom,
            "is_image_attribute": is_image_attribute,
            "is_meta": is_meta,
        }

    def send(self, validated_data) -> ChannelAttributeConfig:
        data = validated_data
        attribute_set = data["attribute_set"]
        attribute = data["attribute"]
        attribute_remote_id = data["attribute_remote_id"]
        is_required = data["is_required"]
        is_variant = data["is_variant"]
        is_custom = data["is_custom"]
        is_image_attribute = data["is_image_attribute"]
        is_meta = data["is_meta"]

        endpoint = self.endpoint(channel_id=self.integration.channel_id)
        try:
            attributeconfig = endpoint.create(
                item=ChannelAttributeConfig(
                    attribute_set=attribute_set,
                    attribute=attribute,
                    attribute_remote_id=attribute_remote_id,
                    is_required=is_required,
                    is_variant=is_variant,
                    is_custom=is_custom,
                    is_image_attribute=is_image_attribute,
                    is_meta=is_meta))
        except HTTPError as e:
            if e.response.status_code // 100 != 4:
                raise e
            attributeconfigs = endpoint.list(
                params={"attribute_set": attribute_set,
                        "attribute": attribute})
            attributeconfig = attributeconfigs[0]
            if is_updated(attributeconfig, data):
                attributeconfig = endpoint.update(
                    id=attributeconfig.pk,
                    item=ChannelAttributeConfig(
                        attribute_set=attribute_set,
                        attribute=attribute,
                        attribute_remote_id=attribute_remote_id,
                        is_required=is_required,
                        is_variant=is_variant,
                        is_custom=is_custom,
                        is_image_attribute=is_image_attribute,
                        is_meta=is_meta))
            return [attributeconfig]
        return [attributeconfig]


class CreateOrUpdateChannelAttributeValue(OmnitronCommandInterface):
    endpoint = ChannelAttributeValueEndpoint

    def get_data(self) -> dict:
        remote_id = self.objects["remote_id"]
        attribute = self.objects["attribute"]
        label = self.objects["label"]
        value = self.objects["value"]
        if not isinstance(remote_id, (str, int)):
            raise AssertionError
        if not isinstance(attribute, int):
            raise AssertionError
        if not isinstance(label, str):
            raise AssertionError
        if not isinstance(value, (str, int)):
            raise AssertionError
        return self.objects

    def send(self, validated_data) -> list:
        # TODO all service calls here can be summarized to a single call to omnitron
        #  which checks IntegrationAction by remote id, if it finds the item, updates the item and
        #  IntegrationAction returns it and halts. If it cannot find the item, it creates the
        #  item, creates its IntegrationAction, returns them and halts.
        data = validated_data
        remote_id = data["remote_id"]
        attribute = data["attribute"]
        label = data["label"]
        value = data["value"]
        channel_id = self.integration.channel_id
        # search in cached attribute values and ia
        attribute_value = self.get_or_create_attribute_value(attribute, label,
                                                             remote_id,
                                                             value, channel_id)
        return [attribute_value]

    @staticmethod
    @lru_cache(maxsize=None)
    def get_or_create_attribute_value(attribute, label, remote_id, value,
                                      channel_id):
        data = {
            "remote_id": remote_id,
            "attribute": attribute,
            "label": label,
            "value": value
        }
        endpoint = ChannelAttributeValueEndpoint(channel_id=channel_id)
        integration_endpoint = ChannelIntegrationActionEndpoint(
            channel_id=channel_id)
        integration_action = integration_endpoint.list(params={
            "content_type_name": "marketplaceattributevalue",
            "channel_id": channel_id,
            "remote_id__exact": remote_id
        })
        attribute_value_item = ChannelAttributeValue(channel=channel_id,
                                                     attribute=attribute,
                                                     label=label,
                                                     value=value, )
        if integration_action:
            integration_action = integration_action[0]
            attribute_value = endpoint.retrieve(
                id=integration_action.object_id)
            if is_updated(attribute_value, data):
                if attribute_value.attribute == data["attribute"]:
                    endpoint.update(id=integration_action.object_id,
                                    item=attribute_value_item)
                else:
                    attribute_value = CreateOrUpdateChannelAttributeValue.create_attribute_value(
                        attribute, attribute_value_item, endpoint, label, value)
        else:
            attribute_value = CreateOrUpdateChannelAttributeValue.create_attribute_value(
                attribute, attribute_value_item, endpoint, label, value)

            content_type = ContentTypeEndpoint().list(params={
                "model": "marketplaceattributevalue"})[0]
            integration_endpoint.create(item=IntegrationAction(
                channel_id=channel_id,
                content_type_id=content_type.id,
                remote_id=remote_id,
                object_id=attribute_value.pk,
                version_date=attribute_value.modified_date,
            ))
        return attribute_value

    @staticmethod
    def create_attribute_value(attribute, attribute_value_item, endpoint, label,
                               value):
        try:
            attribute_value = endpoint.create(item=attribute_value_item)
        except HTTPError:
            attribute_value = endpoint.list(params={"attribute": attribute,
                                                    "label__exact": label,
                                                    "value__exact": value, })[
                0]
        return attribute_value


class GetOrCreateChannelAttributeValueConfig(OmnitronCommandInterface):
    """
    If a subset of attribute values are defined for an attribute set, you need to create
    AttributeValueConfig objects to achieve that setup
    """
    # TODO what happens it a value is no more needed in an attribute set
    #  there is nothing that handles deletes right now

    endpoint = ChannelAttributeValueConfigEndpoint

    def get_data(self) -> dict:
        attribute_set = self.objects["attribute_set"]
        attribute_value = self.objects["attribute_value"]
        if not isinstance(attribute_set, int):
            raise AssertionError
        if not isinstance(attribute_value, int):
            raise AssertionError
        return self.objects

    def send(self, validated_data) -> list:
        data = validated_data
        attribute_set = data["attribute_set"]
        attribute_value = data["attribute_value"]
        channel_id = self.integration.channel_id
        attribute_value_config = self.get_or_create_attribute_value_config(
            attribute_set, attribute_value, channel_id)
        return [attribute_value_config]

    @staticmethod
    @lru_cache(maxsize=None)
    def get_or_create_attribute_value_config(attribute_set, attribute_value,
                                             channel_id):
        endpoint = ChannelAttributeValueConfigEndpoint(channel_id=channel_id)
        attributevalueconfigs = endpoint.list(
            params={"attribute_set": attribute_set,
                    "attribute_value": attribute_value})
        if len(attributevalueconfigs) > 0:
            return attributevalueconfigs[0]
        attribute_value_config = endpoint.create(
            item=ChannelAttributeValueConfig(
                attribute_set=attribute_set,
                attribute_value=attribute_value))
        return attribute_value_config


class GetOrCreateChannelAttributeSchema(OmnitronCommandInterface):
    """
    Schema(Ruleset) objects define how an attribute of a product will be populated. It is
    not feasible to automatically fill the rules. We need to create empty schema objects so that
    Omnitron users can fill the rules
    """

    endpoint = ChannelAttributeSchemaEndpoint

    def get_data(self) -> object:
        return self.objects

    def send(self, validated_data) -> object:
        data = validated_data
        schema = data.get("schema", {})
        channel_id = self.integration.channel_id
        name = f"{str(channel_id)} {data['name']}"
        endpoint = self.endpoint(channel_id=channel_id)
        try:
            attribute_schema = endpoint.create(item=ChannelAttributeSchema(
                name=name, schema=schema, channel=channel_id))
        except HTTPError as e:
            if e.response.status_code // 100 != 4:
                raise e
            attribute_schemas = endpoint.list(params={"name__exact": name,
                                                      "channel": channel_id})
            attribute_schema = attribute_schemas[0]
        return [attribute_schema]


class UpdateChannelConfSchema(OmnitronCommandInterface):
    endpoint = ChannelEndpoint

    def get_data(self) -> dict:
        schema_dict = self.objects
        return schema_dict

    def send(self, validated_data) -> list:
        channel = self.integration.channel
        schema = validated_data
        if isinstance(channel.schema, dict):
            channel.schema.update(schema)
        else:
            channel.schema = schema
        new_channel = Channel(schema=channel.schema)
        channel_response = self.endpoint(
            channel_id=self.integration.channel_id).update(
            id=channel.pk,
            item=new_channel)
        return [channel_response]



from typing import List
from unittest.mock import patch, MagicMock
from omnisdk.base_client import BaseClient
from channel_app.core.data import BatchRequestResponseDto
from channel_app.core.tests import BaseTestCaseMixin
from channel_app.omnitron.batch_request import ClientBatchRequest
from channel_app.omnitron.commands.product_prices import ProcessPriceBatchRequests
from channel_app.omnitron.constants import BatchRequestStatus


class TestProcessPriceBatchRequests(BaseTestCaseMixin):
    """
    Test case for ProcessPriceBatchRequests
    
    run: python -m unittest channel_app.omnitron.commands.tests.test_product_stocks.TestProcessPriceBatchRequests
    """

    def setUp(self) -> None:
        self.instance = ProcessPriceBatchRequests(
            integration=self.mock_integration
        )
        self.instance.objects: List[BatchRequestResponseDto] = [
            BatchRequestResponseDto(
                status='fail',
                remote_id='',
                sku='1',
                message='Error message'
            ),
            BatchRequestResponseDto(
                status='success',
                remote_id='123',
                sku='2',
                message=''
            ),
        ]
        self.get_products_response = {
            "1": MagicMock(
                sku="1"
            ),
            "2": MagicMock(
                sku="3"
            ),
        }

    def test_get_data(self):
        data = self.instance.get_data()
        self.assertEqual(len(data), 2)
        self.assertIsInstance(data[0], BatchRequestResponseDto)
        self.assertIsInstance(data[1], BatchRequestResponseDto)

    def test_validated_data(self):
        data = self.instance.validated_data(self.instance.objects)
        self.assertEqual(len(data), 2)
        self.assertIsInstance(data[0], BatchRequestResponseDto)
        self.assertIsInstance(data[1], BatchRequestResponseDto)

    def test_validated_data_without_valid_data(self):
        self.instance.objects = [{'test': 'test'}]
        
        with self.assertRaises(AssertionError):
            self.instance.validated_data(self.instance.objects)

    def test_update_state(self):
        data = self.instance.update_state
        self.assertEqual(data, BatchRequestStatus.done)

    @patch.object(BaseClient, 'get_instance')
    @patch.object(ClientBatchRequest, 'to_fail')
    def test_check_run(
        self,
        mock_client_batch_request, 
        mock_get_instance
    ):
        result = self.instance.check_run(False, self.instance.objects)
        self.assertFalse(result, False)

    def test_check_run_without_condition(self):
        result = self.instance.check_run(True, self.instance.objects)
        self.assertFalse(result, False)

    @patch.object(BaseClient, 'get_instance')
    @patch.object(ProcessPriceBatchRequests, 'get_products')
    def test_get_channel_items_by_reference_object_ids(
        self,
        mock_get_products,
        mock_get_instance
    ):
        mock_get_products.return_value = self.get_products_response
        data = {
            "productprice": [
                "1"
            ]
        }
        self.instance.integration.channel.conf = {
            "remote_id_attribute": None
        }
        result = self.instance.get_channel_items_by_reference_object_ids(
            self.instance.objects,
            data,
            None
        )
        self.assertEqual(len(result), 1)
        self.assertEqual(result.get("1").sku, "1")



from typing import List

from omnisdk.omnitron.models import ProductPrice, ProductStock, BatchRequest

from channel_app.core import settings
from channel_app.core.data import BatchRequestResponseDto, ErrorReportDto
from channel_app.core.settings import OmnitronIntegration, ChannelIntegration
from channel_app.omnitron.batch_request import ClientBatchRequest
from channel_app.omnitron.constants import ContentType


class PriceService(object):
    batch_service = ClientBatchRequest

    def update_product_prices(self, is_sync=True, is_success_log=True,
                              add_product_objects=False, add_stock=False):
        with OmnitronIntegration(
                content_type=ContentType.product_price.value) as omnitron_integration:
            product_prices = omnitron_integration.do_action(
                key='get_updated_prices')
            first_product_price_count = len(product_prices)

            if add_product_objects:
                product_prices = product_prices and omnitron_integration.do_action(
                    key='get_product_objects', objects=product_prices)

            if add_stock:
                product_prices = product_prices and omnitron_integration.do_action(
                    key='get_stocks_from_product_prices',
                    objects=product_prices,
                    stock_list=omnitron_integration.catalog.stock_list)

            if not product_prices:
                if first_product_price_count:
                    omnitron_integration.batch_request.objects = None
                    self.batch_service(omnitron_integration.channel_id).to_fail(
                        omnitron_integration.batch_request
                    )
                return

            product_prices: List[ProductPrice]

            response_data, reports, data = ChannelIntegration().do_action(
                key='send_updated_prices',
                objects=product_prices,
                batch_request=omnitron_integration.batch_request,
                is_sync=is_sync)

            # tips
            response_data: List[BatchRequestResponseDto]
            reports: List[ErrorReportDto]
            data: List[ProductPrice]

            if not is_sync:
                if reports[0].is_ok:
                    self.batch_service(
                        settings.OMNITRON_CHANNEL_ID).to_sent_to_remote(
                        batch_request=omnitron_integration.batch_request)
                else:
                    is_sync = True

            if reports and (is_success_log or not reports[0].is_ok):
                for report in reports:
                    omnitron_integration.do_action(
                        key='create_error_report',
                        objects=report)

            if is_sync:
                omnitron_integration.do_action(
                    key='process_price_batch_requests',
                    objects=response_data)

    def insert_product_prices(self, is_sync=True, is_success_log=True,
                              add_product_objects=False, add_stock=False):
        with OmnitronIntegration(
                content_type=ContentType.product_price.value) as omnitron_integration:
            product_prices = omnitron_integration.do_action(
                key='get_inserted_prices')
            first_product_price_count = len(product_prices)

            if add_product_objects:
                product_prices = product_prices and omnitron_integration.do_action(
                    key='get_product_objects', objects=product_prices)

            if add_stock:
                product_prices = product_prices and omnitron_integration.do_action(
                    key='get_stocks_from_product_prices',
                    objects=product_prices,
                    stock_list=omnitron_integration.catalog.stock_list)

            if not product_prices:
                if first_product_price_count:
                    omnitron_integration.batch_request.objects = None
                    self.batch_service(omnitron_integration.channel_id).to_fail(
                        omnitron_integration.batch_request
                    )
                return

            product_prices: List[ProductPrice]

            response_data, reports, data = ChannelIntegration().do_action(
                key='send_inserted_prices',
                objects=product_prices,
                batch_request=omnitron_integration.batch_request,
                is_sync=is_sync)

            # tips
            response_data: List[BatchRequestResponseDto]
            reports: List[ErrorReportDto]
            data: List[ProductPrice]

            if not is_sync:
                if reports[0].is_ok:
                    self.batch_service(
                        settings.OMNITRON_CHANNEL_ID).to_sent_to_remote(
                        batch_request=omnitron_integration.batch_request)
                else:
                    is_sync = True

            if reports and (is_success_log or not reports[0].is_ok):
                for report in reports:
                    omnitron_integration.do_action(
                        key='create_error_report',
                        objects=report)

            if is_sync:
                omnitron_integration.do_action(
                    key='process_price_batch_requests',
                    objects=response_data)

    def insert_product_prices_from_extra_price_list(self, is_sync=True,
                                                    is_success_log=True,
                                                    add_stock=False,
                                                    add_product_objects=False):
        currency_mappings = self.get_currency_mappings()
        for price_list_id, country_code in currency_mappings.items():
            with OmnitronIntegration(
                    content_type=ContentType.product_price.value) as omnitron_integration:
                product_prices = omnitron_integration.do_action(
                    key='get_inserted_prices_from_extra_price_list',
                    objects=price_list_id)
                first_product_price_count = len(product_prices)

                if add_product_objects:
                    product_prices = product_prices and omnitron_integration.do_action(
                        key='get_product_objects', objects=product_prices)

                if add_stock:
                    product_prices = product_prices and omnitron_integration.do_action(
                        key='get_stocks_from_product_prices',
                        objects=product_prices,
                        stock_list=omnitron_integration.catalog.stock_list)

                product_prices: List[ProductPrice]
                if product_prices:

                    response_data, reports, data = ChannelIntegration().do_action(
                        key='send_inserted_prices',
                        objects=(product_prices, country_code),
                        batch_request=omnitron_integration.batch_request,
                        is_sync=is_sync)

                    # tips
                    response_data: List[BatchRequestResponseDto]
                    reports: List[ErrorReportDto]
                    data: List[ProductPrice]

                    if not is_sync:
                        if reports[0].is_ok:
                            self.batch_service(
                                settings.OMNITRON_CHANNEL_ID).to_sent_to_remote(
                                batch_request=omnitron_integration.batch_request)
                        else:
                            is_sync = True

                    if reports and (is_success_log or not reports[0].is_ok):
                        for report in reports:
                            omnitron_integration.do_action(
                                key='create_error_report',
                                objects=report)

                    if is_sync:
                        omnitron_integration.do_action(
                            key='process_price_batch_requests',
                            objects=response_data)
                else:
                    if first_product_price_count:
                        omnitron_integration.batch_request.objects = None
                        self.batch_service(omnitron_integration.channel_id).to_fail(
                            omnitron_integration.batch_request
                        )

    def update_product_prices_from_extra_price_list(self, is_sync=True,
                                                    is_success_log=True,
                                                    add_stock=False,
                                                    add_product_objects=False):
        currency_mappings = self.get_currency_mappings()
        for price_list_id, country_code in currency_mappings.items():
            with OmnitronIntegration(
                    content_type=ContentType.product_price.value) as omnitron_integration:
                product_prices = omnitron_integration.do_action(
                    key='get_updated_prices_from_extra_price_list',
                    objects=price_list_id)
                first_product_price_count = len(product_prices)
                if add_product_objects:
                    product_prices = product_prices and omnitron_integration.do_action(
                        key='get_product_objects', objects=product_prices)

                if add_stock:
                    product_prices = product_prices and omnitron_integration.do_action(
                        key='get_stocks_from_product_prices',
                        objects=product_prices,
                        stock_list=omnitron_integration.catalog.stock_list)

                product_prices: List[ProductPrice]
                if product_prices:

                    response_data, reports, data = ChannelIntegration().do_action(
                        key='send_updated_prices',
                        objects=(product_prices, country_code),
                        batch_request=omnitron_integration.batch_request,
                        is_sync=is_sync)

                    # tips
                    response_data: List[BatchRequestResponseDto]
                    reports: List[ErrorReportDto]
                    data: List[ProductPrice]

                    if not is_sync:
                        if reports[0].is_ok:
                            self.batch_service(
                                settings.OMNITRON_CHANNEL_ID).to_sent_to_remote(
                                batch_request=omnitron_integration.batch_request)
                        else:
                            is_sync = True

                    if reports and (is_success_log or not reports[0].is_ok):
                        for report in reports:
                            omnitron_integration.do_action(
                                key='create_error_report',
                                objects=report)

                    if is_sync:
                        omnitron_integration.do_action(
                            key='process_price_batch_requests',
                            objects=response_data)
                else:
                    if first_product_price_count:
                        omnitron_integration.batch_request.objects = None
                        self.batch_service(omnitron_integration.channel_id).to_fail(
                            omnitron_integration.batch_request
                        )
    def get_price_batch_requests(self, is_success_log=True):
        with OmnitronIntegration(create_batch=False) as omnitron_integration:
            batch_request_data = omnitron_integration.do_action(
                'get_batch_requests',
                params={
                    "status": ["sent_to_remote", "ongoing"],
                    "content_type": ContentType.product_price.value})
            # tips
            batch_request_data: List[BatchRequest]

            for batch_request in batch_request_data:
                response_data, reports, data = ChannelIntegration().do_action(
                    key='check_prices',
                    objects=batch_request,
                    batch_request=batch_request
                )

                # tips
                response_data: List[BatchRequestResponseDto]
                reports: List[ErrorReportDto]
                data: BatchRequest

                if reports and (is_success_log or not reports[0].is_ok):
                    for report in reports:
                        omnitron_integration.do_action(
                            key='create_error_report',
                            objects=report)

                if response_data:
                    omnitron_integration.batch_request = batch_request
                    omnitron_integration.do_action(
                        key='process_price_batch_requests',
                        objects=response_data)

    def get_currency_mappings(self):
        """
        Process is like below:
         We get key value list of the currency mappings, filter Omnitron
         with key field
        {
            "1": "ae",
            "2": "us"
        }
        :return:
        """
        with OmnitronIntegration(create_batch=False) as integration:
            currency_mappings = integration.channel.conf.get(
                "CURRENCY_MAPPINGS", {})
            price_lists = integration.catalog.extra_price_lists
            if integration.catalog.price_list:
                price_lists.append(integration.catalog.price_list)
            filtered_currency_mappings = {
                price_list: currency_mappings[price_list]
                for price_list in map(str, price_lists)
                if price_list in currency_mappings}
        return filtered_currency_mappings



class AppException(Exception):
    def __init__(self, params=None):
        super(BaseException, self).__init__()
        self.params = params

class IntegrationMappingException(AppException):
    pass


class CountryException(IntegrationMappingException):
    pass


class CityException(IntegrationMappingException):
    pass


class TownshipException(IntegrationMappingException):
    pass


class DistrictException(IntegrationMappingException):
    pass


class AddressException(IntegrationMappingException):
    pass


class OrderException(AppException):
    pass


class CargoCompanyException(AppException):
    pass


class CustomerException(AppException):
    pass



import logging
import traceback
from datetime import datetime
from io import StringIO
from typing import List, Any

from omnisdk.omnitron.models import BatchRequest
from requests import HTTPError, Request, Response

from channel_app.core.data import ErrorReportDto
from channel_app.core.integration import BaseIntegration
from channel_app.omnitron.batch_request import ClientBatchRequest
from channel_app.omnitron.constants import BatchRequestStatus, ContentType
from channel_app.omnitron.exceptions import (AppException, CityException,
                                             TownshipException,
                                             DistrictException)

logger = logging.getLogger(__name__)


class LogService(object):
    pass


class CommandInterface(object):
    def get_data(self) -> object:
        """
        This method fetches the input data for the command.
        """
        raise NotImplementedError()

    def transform_data(self, data) -> object:
        """
        This method can be used to format the input data before it is executed on the run method.
        """
        raise NotImplementedError()

    def validated_data(self, data) -> dict:
        """
        If the input data needs to satisfy some conditions or contain required a parameter, the
        validation is done here.
        """
        return data

    def send(self, validated_data) -> object:
        """
        If the command sends a request using input data to achieve the main object of the
        command, it is recommended to place those operations in this method.

        :param validated_data:
        """
        raise NotImplementedError()


class ChannelCommandInterface(CommandInterface):
    def __init__(self, integration, objects=None, batch_request=None, **kwargs):
        self.objects = objects
        self.integration = integration
        self.batch_request = batch_request
        self.failed_object_list = []
        self.session = integration._session
        self.CHUNK_SIZE = 50
        self.BATCH_SIZE = 100

        for key, value in kwargs.items():
            setattr(self, "param_{}".format(key), value)

    def get_data(self) -> object:
        raise NotImplementedError

    def validated_data(self, data) -> object:
        raise NotImplementedError

    def transform_data(self, data) -> object:
        raise NotImplementedError

    def send_request(self, transformed_data) -> object:
        """
        If the command sends a request using input data to achieve the main object of the
        command, it is recommended to place those operations in this method.
        """
        raise NotImplementedError()

    def normalize_response(self, data, validated_data, transformed_data,
                           response):
        raise NotImplementedError

    def run(self):
        """
        Main flow of the command. `do_action` method of the integration class executes this method.
        This method must also call necessary command interface methods.
        :return: returns to response of the command if it has one.
        """
        data = self.get_data()
        validated_data = self.validated_data(data)
        transformed_data = self.transform_data(validated_data)
        response = self.send_request(transformed_data=transformed_data)
        normalize_data = self.normalize_response(
            data=data,
            validated_data=validated_data,
            transformed_data=transformed_data,
            response=response)
        return normalize_data

    def create_report(self, response):
        if not self.is_batch_request:
            return
        name = self.__class__.__name__
        if isinstance(response, Response):
            raw_request = f"{response.request.method} - " \
            f"{response.request.url} - " \
            f"{response.request.body}"
            raw_response = response.text
        else:
            raw_request = str("")
            raw_response = str(response)
        
        report_list = []
        report = ErrorReportDto(
            action_content_type=ContentType.batch_request.value,
            action_object_id=self.batch_request.pk,
            modified_date=datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
            error_code=f"{self.batch_request.local_batch_id}-Channel-{name}-{datetime.now().timestamp()}",
            error_description=f"{name}",
            raw_request=raw_request,
            raw_response=raw_response
        )
        is_ok = self.check_response_is_ok(response=response)
        if is_ok:
            report.is_ok = True
        else:
            report.is_ok = False
        report_list.append(report)
        for failed_obj in self.failed_object_list:
            raw_report = ErrorReportDto(
                action_content_type=failed_obj[1],
                action_object_id=failed_obj[0].pk,
                modified_date=failed_obj[0].modified_date,
                error_code=f"{self.batch_request.local_batch_id}-{name}-{datetime.now().timestamp()}",
                error_description=f"{name}",
                raw_request="",
                raw_response=f"{failed_obj[0].failed_reason_type}-{failed_obj[2]}",
                is_ok=False
            )
            report_list.append(raw_report)
        return report_list

    def check_response_is_ok(self, response):
        is_response_obj = isinstance(response, Response)
        if is_response_obj and str(response.status_code).startswith("2"):
            return True
        elif not is_response_obj and response:
            return True
        return False

    @property
    def is_batch_request(self) -> bool:
        is_object = getattr(
            self, 'batch_request', None)
        return bool(is_object)


class OmnitronCommandInterface(CommandInterface):
    """
    Commands implement list of steps to achieve a task. A command can contain a simple get request
    or a complicated business logic with lots of validation.

    """
    batch_service = ClientBatchRequest
    content_type = None

    def __init__(self, integration: BaseIntegration, objects: Any = None,
                 batch_request: BatchRequest = None, **kwargs):
        """
        :param integration: Integration object
        :param objects: Input to the command
        :param batch_request: Related batch request to keep track of the whole task

        """
        self.objects = objects
        self.batch_request = batch_request
        self.integration = integration
        self.failed_object_list = []
        for key, value in kwargs.items():
            setattr(self, "param_{}".format(key), value)

    def get_data(self) -> object:
        return self.objects

    def send(self, validated_data) -> object:
        return validated_data

    def normalize_response(self, data, response) -> List[object]:
        return response

    def run(self) -> Any:
        """
        Main flow of the command. `do_action` method of the integration class executes this method.
        This method must also call necessary command interface methods.
        :return: Response of the command or None
        """
        is_ok = True
        formatted_data = None
        raw_request, raw_response = None, None
        try:
            model_items = self.validated_data(self.get_data())
            response = self.send(validated_data=model_items)
            normalize_data = self.normalize_response(data=model_items,
                                                     response=response)
            if isinstance(normalize_data, list):
                formatted_data = [model_obj for model_obj in normalize_data
                                  if
                                  not getattr(model_obj, "failed_reason_type",
                                              None)]
            self.row_send_error_report()
        except HTTPError as e:
            request = e.request
            raw_request = f"{request.method} - {request.url} - {request.body}"
            raw_response = e.response.text
            is_ok = False
            logger.error(f"{raw_request}-/-{raw_response}")
        except (CityException, TownshipException, DistrictException) as e:
            is_ok = False
            self.integration.do_action(
                key='create_address_error_report',
                objects=e.params)
            raw_response = str(e.params)
        except AppException as e:
            is_ok = False
            raw_response = str(e.params)
        except Exception as e:
            is_ok = False
            raw_response = f"{str(e)} - {traceback.format_exc()}"
            request = getattr(e, "request", "")
            if request:
                raw_request = f"{request.method} - {request.url} - {request.body}"
            logger.error(f"{raw_request}-/-{raw_response}")
        finally:
            if not is_ok:
                self.send_error_report(raw_request, raw_response)

        is_check_and_update_batch_service = self.check_run(
            is_ok=is_ok, formatted_data=formatted_data)
        if not is_check_and_update_batch_service:
            return []

        return formatted_data

    def check_run(self, is_ok, formatted_data):
        if is_ok and not \
            (formatted_data or self.failed_object_list) and \
            self.is_batch_request:
            self.batch_service(self.integration.channel_id).to_done(
                self.integration.batch_request)
            return False
        elif not is_ok and self.is_batch_request:
            self.integration.batch_request.objects = None
            self.batch_service(self.integration.channel_id).to_fail(
                self.integration.batch_request)
            return False
        return True

    def row_send_error_report(self):
        name = self.__class__.__name__
        for failed_obj in self.failed_object_list:
            report = ErrorReportDto(
                action_content_type=failed_obj[1],
                action_object_id=failed_obj[0].pk,
                modified_date=failed_obj[0].modified_date,
                error_code=f"{self.integration.batch_request.local_batch_id}-Omnitron-{name}",
                error_description=f"Omnitron-{name}",
                raw_request="",
                raw_response=f"{failed_obj[0].failed_reason_type}-{failed_obj[2]}",
                is_ok=False
            )
            self.integration.do_action(key='create_error_report',
                                       objects=report)

    def send_error_report(self, raw_request, raw_response):
        if not self.is_batch_request:
            return

        name = self.__class__.__name__

        report = ErrorReportDto(
            action_content_type=ContentType.batch_request.value,
            action_object_id=self.integration.batch_request.pk,
            modified_date=datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
            error_code=f"{self.integration.batch_request.local_batch_id}-{name}-{datetime.now().microsecond}",
            error_description=f"Omnitron-{name}",
            raw_request=raw_request,
            raw_response=raw_response
        )
        self.integration.do_action(
            key='create_error_report',
            objects=report)

    @property
    def is_batch_request(self) -> bool:
        is_object = getattr(
            self.integration, 'batch_request', None)
        return bool(is_object)

    def create_log_file(self, message: str = None, raw_request: Request = None,
                        raw_response: Response = None, **kwargs):
        """
        Log files are attached to batch requests. If you need to attach logs
        to a problematic command or want to see detailed logs for standard a
        command, you can use this method. However, if you do not use a batch request
        for a command and still want to see the logs, You will need to manage the command
        with a batch request.

        :param message: Reason of the exception/entry logged
        :param raw_request:
        :param raw_response:
        """
        bytes_io = StringIO()
        if raw_request and raw_response:
            context = f"""
            "message": {message},
            "url": {raw_request.url},
            "request_body": {raw_request.body},
            "response_text": {raw_response.text},
            "response_reason": {raw_response.reason},
            "response_status_code": {raw_response.status_code}
            """
        else:
            context = f"""
                        "message": {message}
                        """
        bytes_io.write(context)
        bytes_io.seek(0)
        return bytes_io

    @property
    def update_state(self, *args, **kwargs) -> BatchRequestStatus:
        return BatchRequestStatus.commit

    def update_batch_request(self, objects_data: list):
        """
        Batch requests are used to track state of long-running processes across multiple
        systems. State of those requests can be managed through this method.

        :param content_type: String values of the ContentType enum model
        :param remote_batch_id: Batch request id of the
        :param objects_data:

        """
        self.integration.batch_request.objects = objects_data
        service = self.batch_service(self.integration.channel_id)
        state = getattr(service, f'to_{self.update_state.value}')(
            self.integration.batch_request)
        return state

    def create_batch_objects(self, data: list, content_type: str) -> List[dict]:
        """
        In batch requests, you can attach Omnitron objects that are being processed in that batch
        so that they are excluded from further calls until that batch is finalized with done or
        fail statuses.

        :param data: List of objects which are being processed for this batch request
        :param content_type: String values of the ContentType enum model

        """
        objects = []
        for item in data:
            object = {
                "pk": item.pk,
                "failed_reason_type": getattr(item, "failed_reason_type", None),
                "remote_id": getattr(item, "remote_id", None),
                "version_date": item.modified_date,
                "content_type": content_type
            }
            objects.append(object)
        return objects
